////////////////////////////////////////////////////////////////////
// Netlist.cc: Definition of Netlist member functions.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

#include "Netlist.h"

// Set name of blif netlist to n.
void Netlist::set_name(char* n){
  if(!n || strlen(n)>MAX_STR_LEN)
    ERROR("error in naming netlist...");
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  strcpy(name,n);
}

// Add p_clb to list of logic blocks.
// This function assumes that p_clb has been updated
// to reflect whether it is a logic, input-only or
// output-only block. 
void Netlist::addto_lbList(logicBlock* p_clb){
  if(!p_clb)
    ERROR("logic block cannot be NULL...");
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  lbList.push_back(p_clb);

  if(p_clb->get_isLogic())
    numlogicBlocks++;
  else if(p_clb->get_isInput())
    numInputs++;
  else if(p_clb->get_isOutput())
    numOutputs++;
}

// for each net in this netlist, create vector of sinks, and then populate
void Netlist::create_netSinkVectors(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  list<Net*>::iterator net_itr = netList.begin();

  for(;net_itr!=netList.end();net_itr++)
    (*net_itr)->create_sinkVector();

}

// for each net in this netlist, sort the sinks in increasing order of separation from source
void Netlist::sort_netSinks(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");

  list<Net*>::iterator net_itr = netList.begin();

  for(;net_itr!=netList.end();net_itr++){
    if((*net_itr)->sinkVector)
      (*net_itr)->sort_sinkVector();
  }

}

// Create vector of pointers to logic blocks that comprise this netlist. 
// Note that this function can be called only AFTER lbList has been completely populated.
void Netlist::create_lbVector(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  int num_elements=numlogicBlocks+numInputs+numOutputs;
  lbVector=new vector<logicBlock*>(num_elements);
  list<logicBlock*>::iterator lb_itr = lbList.begin();
  for(int i=0;lb_itr!=lbList.end();lb_itr++,i++)
    lbVector->at(i)=*lb_itr;
}

// Return logic block at lbVector[i]
logicBlock* Netlist::get_logicBlock(int i){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  if(i >= lbVector->size())
    ERROR("lbVector overflow here...");
  return lbVector->at(i);
}

// Add p_net to list of nets and increment num_nets.
void Netlist::addto_netList(Net* p_net){
  if(!p_net)
    ERROR("net cannot be NULL...");
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  netList.push_back(p_net);
  numNets++;
}

// Get total num logic blocks in this netlist.
int Netlist::get_numlogicBlocks(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  return numlogicBlocks;
}

// Get total num input-only IO blocks in this netlist.
int Netlist::get_numInputs(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  return numInputs;
}

// Get total num output-only IO blocks in this netlist.
int Netlist::get_numOutputs(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  return numOutputs;
}

// Get total num nets in this netlist.
int Netlist::get_numNets(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  return numNets;
}

// Get the name of this netlist
char* Netlist::get_name(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  return name;
}

// Print details of each logic block in lbList.
void Netlist::print_lbList(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");

  cout<<endl<<endl<<"***************************************************************";
  cout<<endl<<"\t\tBEGIN PRINT LOGIC BLOCK LIST";
  cout<<endl<<"***************************************************************";

  list<logicBlock*>::iterator lb_itr = lbList.begin();
  for(;lb_itr!=lbList.end();lb_itr++)
    (*lb_itr)->print_Info();

  cout<<endl<<endl<<"***************************************************************";
  cout<<endl<<"\t\tEND PRINT LOGIC BLOCK LIST";
  cout<<endl<<"***************************************************************";

}

// Print details of each net in netList.
void Netlist::print_netList(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");

  cout<<endl<<endl<<"***************************************************************";
  cout<<endl<<"\t\tBEGIN PRINT NET LIST";
  cout<<endl<<"***************************************************************";

  list<Net*>::iterator net_itr = netList.begin();
  for(;net_itr!=netList.end();net_itr++)
    (*net_itr)->print_Info();

  cout<<endl<<endl<<"***************************************************************";
  cout<<endl<<"\t\tEND PRINT NET LIST";
  cout<<endl<<"***************************************************************";

}

// Print the name of this netlist.
void Netlist::print_name(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  cout<<endl<<endl<<"NAME OF NETLIST IS "<<name;
}

// Print the details of logic block pointed to by p_clb.
void Netlist::print_logicBlock(logicBlock* p_clb){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  if(!p_clb)
    ERROR("logic block cannot be NULL...");
  p_clb->print_Info();
}

// Print the details of net pointed to by p_net.
void Netlist::print_Net(Net* p_net){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  if(!p_net)
    ERROR("net cannot be NULL...");
  p_net->print_Info();
}

// Look for logic block with clb_name in lbList. If found, return the logic block.
logicBlock* Netlist::found_logicBlock(char* clb_name){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  if(!clb_name)
    ERROR("logic block name cannot be NULL...");
  list<logicBlock*>::iterator lb_itr = lbList.begin();
  for(;lb_itr!=lbList.end();lb_itr++){
    if(strcmp((*lb_itr)->get_name(),clb_name)==0)
      return *lb_itr;
  }
  return 0;
}

// Look for net with net_name in netList. If found, return the net.
Net* Netlist::found_Net(char* net_name){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  if(!net_name)
    ERROR("net name cannot be NULL...");
  list<Net*>::iterator net_itr = netList.begin();
  for(;net_itr!=netList.end();net_itr++){
    if(strcmp((*net_itr)->get_name(),net_name)==0)
      return *net_itr;
  }
  return 0;
}

// Create a new iterator for lbList
void Netlist::create_lbListItr(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  if(lb_itr)
    ERROR("logic block list iterator already exists...");
  lb_itr=new list<logicBlock*>::iterator;
  *lb_itr = lbList.begin();
}

// Remove an already existing lbList iterator 
void Netlist::remove_lbListItr(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  if(lb_itr==0)
    ERROR("trying to delete NULL logic block list iterator...");
  delete lb_itr;
  lb_itr=0;
}

// Create a new iterator for netList
void Netlist::create_netListItr(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  if(net_itr)
    ERROR("net list iterator already exists...");
  net_itr=new list<Net*>::iterator;
  *net_itr = netList.begin();
}

// Remove an already existing netList iterator 
void Netlist::remove_netListItr(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  if(net_itr==0)
    ERROR("trying to delete NULL net iterator...");
  delete net_itr;
  net_itr=0;
}

// Get next element from lbList
logicBlock* Netlist::getnext_lbList(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");

  if(*lb_itr == lbList.end())
    return 0;
  else{
    logicBlock* ret_ptr = **lb_itr;
    (*lb_itr)++;
    return ret_ptr;
  }

}

// Get next element from netList
Net* Netlist::getnext_netList(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");

  if(*net_itr == netList.end())
    return 0;
  else{
    Net* ret_ptr = **net_itr;
    (*net_itr)++;
    return ret_ptr;
  }

}


