////////////////////////////////////////////////////////////////////
// Netlist.h: Class that describes a generic Netlist. This is
//            where a netlist in some given format is converted
//            to a netlist that can be used by Independence.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

/****************************************************************
Class definition for a generic netlist. A netlist consists of
a list of all logic blocks in the netlist, and a separate list
of the nets that connect the logic blocks. Further, we also keep
a record of the number of IO blocks, logic blocks and nets in 
the netlist.
****************************************************************/ 
#ifndef NETLIST_H
#define NETLIST_H

#include "error.h"
#include "logicBlock.h"
#include "Net.h"

class Netlist{
  char name [MAX_STR_LEN];  //name of netlist
  list<logicBlock*> lbList;  //list of logic blocks in netlist
  list<Net*> netList;  //list of nets in netlist
  int numlogicBlocks;  //num logic blocks in netlist
  int numInputs;  //num input-only IO blocks
  int numOutputs;  //num output-only IO blocks
  int numNets;  //num nets in netlist

  vector<logicBlock*>* lbVector;  //pointer to vector of logic blocks in this netlist

  list<logicBlock*>::iterator* lb_itr;  //pointer to lbList iterator
  list<Net*>::iterator* net_itr;  //pointer to netList iterator

  int magicNum;
 public:
  Netlist(){
    numlogicBlocks=0;
    numNets=0;
    numInputs=numOutputs=0;
    lb_itr=0;net_itr=0;
    lbVector=0;
    magicNum=18111;
  }

  ~Netlist(){
    if(magicNum!=18111)
      ERROR("trying to double delete a node...");
    else
      magicNum=0;
  }

  void create_lbVector();  //create vector of pointers to logic blocks that comprise 
                              //this netlist. note that this function can be called
                              //only AFTER lbList has been completely populated.
  logicBlock* get_logicBlock(int i); //return logic block at lbVector[i]
  void create_lbListItr();  //create a new iterator for lbList
  void create_netListItr();  //create a new iterator for netList
  void remove_lbListItr();  //remove lbList iterator
  void remove_netListItr();  //remove netList iterator
  logicBlock* getnext_lbList();  //get next element from lbList
  Net* getnext_netList();  //get next element from netList

  void set_name(char* n);  //set the name of netlist
  void addto_lbList(logicBlock* p_clb);  //add a logic block to lbList
  void addto_netList(Net* p_net);  //add a net to netList

  int get_numlogicBlocks();  //get num logic blocks in netlist
  int get_numNets();  //get num nets in netlist
  char* get_name();  //get the name of blif netlist
  int get_numInputs();  //get num input-only IO blocks
  int get_numOutputs();  //get num output-only IO blocks

  void print_lbList();  //print the entire list of logic blocks
  void print_netList();  //print the entire list of nets
  void print_name();  //print the name of netlist
  void print_logicBlock(logicBlock* p_clb);  //print details of logic block pointed to 
                                                //by p_clb
  void print_Net(Net* p_net);    //print details of net pointed to by p_net

  logicBlock* found_logicBlock(char* clb_name);  //look for logic block in lbList by name.
  Net* found_Net(char* net_name);  //look for net in netList by name.
  
  // for each net in this netlist, create vector of sinks, and then populate
  void create_netSinkVectors();

  // for each net in this netlist, sort the sinks in increasing order of separation from source  
  void sort_netSinks();
};

#endif


