////////////////////////////////////////////////////////////////////
// hsraArch.h : Classes that describe the HSRA architecture.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

/****************************************************************
This file contains class definitions for the HSRA architecture
and the individual clbs that constitute a HSRA-style FPGA. Each
hsraCLB object holds information about the source/sink/input-pin/
output-pin routing nodes that it provides. The hsraCLB also holds 
a pointer to the logic block that is currently placed on it.
****************************************************************/    
#ifndef HSRA_ARCH_H
#define HSRA_ARCH_H
#include "routingEdge.h"
#include "routingNode.h"

extern double flomax(double x,double y);

class hsraBlock{
 public:
  int position;  // linear position of this block in the HSRA array
  int type;  //whether this is a CLB or IO block
  routingNode* sourceNode;  //source routing node
  routingNode* sinkNode;  //sink routing node
  list<routingNode*> inputPinList;  //list of input routing nodes
  list<routingNode*> outputPinList;  //list of output routing nodes
  list<routingNode*>::iterator* ipin_itr;  //iterator for inputPinList
  list<routingNode*>::iterator* opin_itr;  //iterator for outputPinList
  logicBlock* currBlock;  //the logic block that is currently placed at this CLB
  int magicNum;

  hsraBlock(int p,int t);
  ~hsraBlock();

  int get_position();  //get the position of this block
  int get_type();  //whether this is a CLB or IOB
  void set_currBlock(logicBlock* l);  //set the logic block that is placed at this CLB
  logicBlock* get_currBlock();  //get the logic block that is placed at this CLB
  void set_sourceNode(routingNode* r);  //set the source node of this CLB
  void set_sinkNode(routingNode* r);  //set the sink node of this CLB
  routingNode* get_sourceNode();  //get the source node of this CLB
  routingNode* get_sinkNode();  //get the sink node of this CLB
  void addto_inputPinList(routingNode* r);  //add to list of input routing nodes
  void addto_outputPinList(routingNode* r);  //add to list of output routing nodes
  void create_inputPinListItr();  //create an iterator for inputPinList
  void create_outputPinListItr();  //create an iterator for outputPinList
  void remove_inputPinListItr();  //remove the iterator for inputPinList
  void remove_outputPinListItr();  //remove the iterator for outputPinList
  routingNode* getnext_inputPinList();  //get next routing node in inputPinList
  routingNode* getnext_outputPinList();  //get next routing node in outputPinList
  void print_info();  //print details about this CLB

};

class switchBox{
 public:
  int numSwitches;  //num switches in this switch box
  int switchType;  //type of switch...COMPRESSING or NON_COMPRESSING
  vector<routingNode*>* LArray;  //array of routing nodes that connect to the the 'L' 
                                 //input of each switch...refer DeHon's FPGA '99 papers.
                                 //Each switchbox 'owns' the routing nodes that connect
                                 //to its L and R inputs.
  vector<routingNode*>* RArray;
  int magicNum;

  switchBox(int num_switches,int switch_type);
  ~switchBox(){
    if(magicNum!=102938)
      ERROR("trying to double delete node...");
    else
      magicNum=0;
  }

  int get_numSwitches();  //get num switches in this switchbox
  int get_switchType();  //get the type of switch i.e. COMPRESSING or NON-COMPRESSING
  routingNode* get_LRNodeAt(int index);  //get routing node at LArray[index]
  routingNode* get_RRNodeAt(int index);  //get routing node at RArray[index]
  void addto_LArray(routingNode* r,int index);  //add routing node at LArray[index]
  void addto_RArray(routingNode* r,int index);  //add routing node at RArray[index]

};

class levelStats{
 public:
  int channelWidth;  //channel width at this level
  int switchType;  //type of switch at this level...COMPRESSING or NON_COMPRESSING
  int numSwitchBoxes;  //total number of switch-boxes at this level
  double continuousApprox;  //this is the actual approximation of the interconnect
                              //requirements when going from this level to the next
  vector<switchBox*>* switchArray;  //this array will hold each switchbox at this level
                                    //of the interconnect hierarchy 
  int magicNum;

  levelStats(int chan_width,int switch_type,int num_switch_boxes);
  ~levelStats(){
    if(magicNum!=45678)
      ERROR("trying to double delete node...");
    else
      magicNum=0;
  }

  int get_channelWidth();  //get the channel width at this level
  int get_switchType();  //get the type of switch at this level
  int get_numSwitchBoxes();  //get number of switchboxes at this level
  switchBox* get_switchBoxAt(int index);  //get switch box at switchArray[index] 
  void addto_switchArray(switchBox* sw,int index);  //add switchbox at switchArray[index] 
  void print_info(FILE* fp);

};

class hsraFPGA{
 public:
  int numCLBs,numIOBs;  //num CLBs and IOBs in this architecture
  int baseChanWidth;  //the base channel width of this architecture
  int numLevels; //number of levels in the interconnect tree
  double growthRate; //the rate at which the interconnect grows from leaves to root 
  vector<hsraBlock*>* clbArray;  //linear array of CLBs of size numCLBs
  vector<hsraBlock*>* iobArray;  //linear array of IOBs of size numIOBs
  vector<levelStats*>* levelArray;  //stores info about each tree level
  vector<routingNode*>* ioWireArray;  //the routing nodes that connect IOBs to the interconnect
  list<routingNode*> rnodeList;  //List that holds ALL routing nodes in the architecture
  list<routingNode*>::iterator* rnode_itr;  //iterator for list of routing nodes
  list <routingEdge*> redgeList;  //list of ALL routing edges in the routing graph

  vector<routingNode*>* wire_rnodeVec;
  vector<routingNode*>* sink_rnodeVec;
  vector<routingNode*>* rnodeVec;

  map<string,routingNode*> rnodeMap;

  routingNode* iobSource;  //HSRA style...all IOBs are fed from a single source routing node
  routingNode* iobSink;   //HSRA style...all IOBs connect to a single sink routing node
  int lsize;
  int numSinkNodes;

  int magicNum;

  hsraFPGA(int num_clbs,int base_width,int num_levels,double growth_rate,int lsz);
  ~hsraFPGA();

  hsraBlock* get_clbAt(int pos);  //get clb at position pos
  hsraBlock* get_iobAt(int pos);  //get iob at position pos
  int get_numCLBs();  //get num CLBs in this architecture
  int get_numIOBs();  //get num IOBs in this architecture
  void addto_rnodeList(routingNode* r);  //add routing nodes to rnodeList
  routingNode* getnext_rnodeList();  //get next routing node in rnodeList
  void create_rnodeListItr();  //create an iterator for rnodeList
  void remove_rnodeListItr();  //remove iterator for rnodeList
  routingNode* found_routingNode(char* rname);  //look for routing node with rname and return

  int get_baseChanWidth();  //get the base channel width of the HSRA
  int get_numLevels();  //get the number of levels in the interconnect tree
  double get_growthRate();  //get the interconnect growth rate

  //count (and optionally print) the total sharing of routing nodes in this FPGA
  long int countSharing(int opt);
  long int countSharing(FILE* fp);

  //update the history costs of all routing nodes in this FPGA
  void updateHistoryCosts(int iter,bool during_placement);

  // create the routing nodes that will constitute the interconnect structure
  void create_routingNodes();

  // create the edges that will connect routing nodes
  void create_routingEdges();

  // create a routing edge
  void create_routingEdge(routingNode* orig_node,routingNode* dest_node);

  // populate the LRLists of the input and sink nodes of each CLB
  void populate_CLB_LRLists();

  // populate the LRLists of each wire in the interconnect structure
  void populate_LRList(routingNode* rnode,int level_num,int sbox_num,bool left);

  void print_info(FILE* fp);  //print arch info about this FPGA

};


#endif

