////////////////////////////////////////////////////////////////////
// hsraMain.y:   This is the file in which the yacc grammar resides.
//               Further, all main placement and routing routines
//               are resident here.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

%{
#include "hsraMain.h"
%}

%error-verbose

%union{
  char* word;
  double double_num;
}

%token <word> INPUTS_TOK 
%token <word> OUTPUTS_TOK 
%token <word> MODEL_TOK 
%token <word> SLOPE_TOK 
%token <word> END_TOK 
%token <word> LATCH_TOK
%token <word> NAMES_TOK
%token <double_num> DOUBLENUM_TOK
%token <word> NET_TOK
%token <word> LOGIC_TOK

%token <word> CLUSTER_TOK
%token <word> NUM_CLUSTERS_TOK
%token <word> NUM_SINKS_TOK
%%
statement_list: statement
|    statement_list statement
;

statement: circuit_info{
  stringListItr = stringList.begin();
  for(;stringListItr!=stringList.end();stringListItr++)
    free(*stringListItr);
  stringList.clear();
};
|
cluster_info{
  ;
}
;

cluster_info:
  NUM_CLUSTERS_TOK NET_TOK { 
  numClusters = atoi($2);free($1);free($2); 
}
| NUM_SINKS_TOK NET_TOK { 
  prepareAstarTable(numClusters,atoi($2));free($1);free($2);
/*   exit(1); */
}
| NET_TOK NET_TOK DOUBLENUM_TOK {
  process_AstarTableEntry(atoi($1),$2,$3);
  free($1);free($2);
}
;

circuit_info: 
MODEL_TOK NET_TOK{ 
  free($1);free($2);
}
| INPUTS_TOK nets{
  processBlifInputs();
  free($1);
}
| OUTPUTS_TOK nets{
  processBlifOutputs();
  free($1);
}
| LATCH_TOK NET_TOK NET_TOK NET_TOK{
  processBlifLatch($2,$3);
  free($1);free($2);free($3);free($4);
}
| LATCH_TOK NET_TOK NET_TOK NET_TOK NET_TOK NET_TOK{
  processBlifLatch($2,$3);
  free($1);free($2);free($3);free($4);free($5);free($6);
}
| NAMES_TOK nets{
  processBlifClb();
  free($1);
}
| SLOPE_TOK DOUBLENUM_TOK{
  free($1);
}
| LOGIC_TOK NET_TOK{
  if(READ_PLACEMENT){
    placeVec->at(atoi($1))=$2;
    free($1);
  }
  else{
    free($1);free($2);
  }
}
| END_TOK{
  free($1);
}
;

nets:
|    NET_TOK nets   { stringList.push_front($1); }
;
%%

/* input file that is to be parsed */
extern FILE* yyin;

int main(int argc,char** argv){
  int itr;
  time_t start_time,end_time;
  FILE* fp;char outfile_name [MAX_STR_LEN];

  try{

/*     if(argc<9){ */
/*       cout<<endl<<endl<<"Insufficient command-line parameters. Minimum syntax is -->"; */
/*       cout<<endl<<"./independence <netlist> <num_clbs> <base_chan_width> <num_levels> <growth_rate> <lsize> <cluster_file> <table_file>"; */
/*       cout<<endl<<"<netlist>         : name of netlist in the .4LUT format."; */
/*       cout<<endl<<"<num_clbs>        : number of logic blocks in a target HSRA that is a complete binary tree. Must be a power of 2."; */
/*       cout<<endl<<"<base_chan_width> : number of tracks in base channle of target HSRA. Generally, this is is at least 5."; */
/*       cout<<endl<<"<num_levels>      : number of levels in a target HSRA that is a complete binary tree. Ensure that 2^<num_levels> == <num_clbs>."; */
/*       cout<<endl<<"<growth_rate>     : rate at which interconnect grows towards the root. This needs to be a floating-point num between 0 and 1."; */
/*       cout<<endl<<"<lsize>           : limit the placement to logic blocks between 1...<lsize>. <lsize> should be less-than-equal-to <num_clbs>."; */
/*       cout<<endl<<"<cluster_file>    : name of file that contains the cluster assignments for each routing node"; */
/*       cout<<endl<<"<table_file>      : name of file that contains the Astar table entries for each cluster."; */
/*       cout<<endl<<endl<<"Optional parameters -->"; */
/*       cout<<endl<<"<place_file>      : HSRA-style placement file. Needed when you wanna bypass placement and do only routing."; */
/*       cout<<endl<<endl; */
/*       exit(1); */
/*     } */








    if(argc<6){
      cout<<endl<<endl<<"Insufficient command-line parameters. Minimum syntax is -->";
      cout<<endl<<"./hsra_gen <num_clbs> <base_chan_width> <num_levels> <growth_rate> <lsize>";
      cout<<endl<<"<num_clbs>        : number of logic blocks in a target HSRA that is a complete binary tree. Must be a power of 2.";
      cout<<endl<<"<base_chan_width> : number of tracks in base channel of target HSRA. Generally, this is is at least 5.";
      cout<<endl<<"<num_levels>      : number of levels in a target HSRA that is a complete binary tree. Ensure that 2^<num_levels> == <num_clbs>.";
      cout<<endl<<"<growth_rate>     : rate at which interconnect grows towards the root. This needs to be a floating-point num between 0 and 1.";
      cout<<endl<<"<lsize>           : limit the placement to logic blocks between 1...<lsize>. <lsize> should be less-than-equal-to <num_clbs>.";
      cout<<endl<<endl;
      exit(1);
    }

    /* create the HSRA architecture */
    create_hsraFPGA(atoi(argv[1]),atoi(argv[2]),atoi(argv[3]),atof(argv[4]),atoi(argv[5]));

    /* dump the HSRA architecture */
    FILE* rrgraph_fp;
    int nclbs=FPGA->get_numCLBs();int base_chan=FPGA->get_baseChanWidth();
    double growth=FPGA->get_growthRate();
    char rrgraph_name [MAX_STR_LEN];
    sprintf(rrgraph_name,"%d_%d_%.3f.hsra",nclbs,base_chan,growth);
    rrgraph_fp=fopen(rrgraph_name,"w");
    dumpArchInRRGraph(rrgraph_fp);
    fclose(rrgraph_fp);
    exit(1);








    /* open input file */
    yyin=fopen(argv[1],"r");

    randomSeed=time(NULL);
    srand48(randomSeed);

    /* give the mcnc netlist its name */
    blif_netlist.set_name(argv[1]);

    /* parse input file */
    while(!feof(yyin)){
      yyparse();
    }

    /* close input file */
    fclose(yyin);

    /* convert blif to generic internal netlist */
    convert2GenericNetlist(blif_netlist);

    numBlocks=netlist.get_numlogicBlocks();
    numNets=netlist.get_numNets();

    readInClusterAssignments(argv[7]);

/*     dumpClusterAssignments(); */
/*     exit(1); */

    /* read in table entries */
    yyin=fopen(argv[8],"r");
    while(!feof(yyin)){
      yyparse();
    }
    fclose(yyin);

/*     dumpAstarTable(); */
/*     exit(1); */

/*     slogEstimates(); */

    sprintf(outfile_name,"%s_%s.out",netlist.get_name(),argv[7]);
    fp=fopen(outfile_name,"w");

    cout<<endl<<endl<<"Building AllSinksTables....";
    fprintf(fp,"\n\nBuilding AllSinksTables....");
    start_time=time(0);
    populateAllSinksTables();
    end_time=time(0);
    cout<<"done!!";
    fprintf(fp,"done!!");
    cout<<" Total time to build AllSinksTables is "<<(end_time-start_time)<<endl;
    fprintf(fp," Total time to build AllSinksTables is %d\n",end_time-start_time);

/*     dumpCostToTargetEstimates(); */

    performAstarSanityCheck(fp);

/*     dumpAllSinksTables(); */
/*     exit(1); */

    if(argc==10){

      cout<<endl<<endl;
      cout<<endl<<"Netlist is "<<argv[1];
      cout<<endl<<"Total num blocks (LOGIC) = "<<numBlocks<<" and total num nets = "<<numNets;
      cout<<endl<<"Target is an HSRA with "<<FPGA->get_numCLBs()<<" logic blocks";
      cout<<endl<<"Target LSIZE is "<<FPGA->lsize;
      cout<<endl<<"The base channel width is "<<FPGA->get_baseChanWidth()<<" and growth rate is "<<FPGA->get_growthRate();
      cout<<endl;

      fprintf(fp,"\n\n\nNetlist is %s",argv[1]);
      fprintf(fp,"\nTotal num blocks (LOGIC) is %d, and total num nets is %d",numBlocks,numNets);
      fprintf(fp,"\nTarget is an HSRA with %d logic blocks",FPGA->get_numCLBs());
      fprintf(fp,"\nTarget LSIZE is %d",FPGA->lsize);
      fprintf(fp,"\nThe base channel width is %d, and growth rate is %f",FPGA->get_baseChanWidth(),FPGA->get_growthRate());
      fprintf(fp,"\n");

      placeVec=new vector<char*>(atoi(argv[2]));
      for(int i=0;i<placeVec->size();i++)
	placeVec->at(i)=0;

      /* read in placement */
      yyin=fopen(argv[9],"r");

      READ_PLACEMENT=true;

      /* parse placement file */
      while(!feof(yyin)){
	yyparse();
      }
      /* close placement file */
      fclose(yyin);

      READ_PLACEMENT=false;

      initExternalPlacement();

      /* to facilitate A*, sort sinks in increasing order of separation from source */
      netlist.sort_netSinks();

      routeNets(1,true);
      sharingCost=FPGA->countSharing(0);wireCost=calcWireCost();

      /* start measuring the routing runtime */
      start_time=time(0);

      cout<<endl<<endl<<"<<<<<<<<<<<<<<<<<<<<<< BEGIN ROUTING PHASE >>>>>>>>>>>>>>>>>>>>>>>>";
      fprintf(fp,"\n\n<<<<<<<<<<<<<<<<<<<<<< BEGIN ROUTING PHASE >>>>>>>>>>>>>>>>>>>>>>>>");
      cout<<endl<<"ITERATION 1\tSHARING = "<<sharingCost;
      fprintf(fp,"\nITERATION 1\tSHARING = %d",sharingCost);
      for(itr=2;itr<=200;itr++){
	FPGA->updateHistoryCosts(itr,false);
	routeNets(itr,true);
	cout<<endl<<"ITERATION "<<itr<<"\tSHARING = "<<sharingCost;
	fprintf(fp,"\nITERATION %d\tSHARING = %d",itr,sharingCost);
	if(sharingCost==0){
	  cout<<endl<<endl<<"SUCCESFULLY PLACED AND ROUTED THIS BLOKE!!";
	  cout<<endl<<endl<<"<<<<<<<<<<<<<<<<<<<<<< END ROUTING PHASE >>>>>>>>>>>>>>>>>>>>>>>>";
	  cout<<endl<<endl;

	  fprintf(fp,"\n\nSUCCESSFULLY PLACED AND ROUTED THIS BLOKE!!");
	  fprintf(fp,"\n\n<<<<<<<<<<<<<<<<<<<<<< END ROUTING PHASE >>>>>>>>>>>>>>>>>>>>>>>>\n\n");

	  break;
	}
      }

      /* end measuring the routing runtime */
      end_time=time(0);

      cout<<endl<<endl<<"TOTAL ROUTING RUNTIME IS "<<(end_time-start_time);
      fprintf(fp,"\n\nTOTAL ROUTING RUNTIME IS %d\n",end_time-start_time);
      cout<<endl;

      cout<<endl<<"RUNTIME PER ROUTING ITERATION IS "<<((float)(end_time-start_time))/(float)(itr);
      cout<<endl;
      fprintf(fp,"\nRUNTIME PER ROUTING ITERATION IS %f\n",((float)(end_time-start_time))/(float)(itr));

      if(itr==201){
	printNetsAndRecalcCosts();
	printNetsToFile(fp);
	cout<<endl<<endl<<"CLOSE, BUT NO CIGAR....";
	cout<<endl<<endl<<"<<<<<<<<<<<<<<<<<<<<<< END ROUTING PHASE >>>>>>>>>>>>>>>>>>>>>>>>";
	cout<<endl<<endl;

	fprintf(fp,"\n\nCLOSE, BUT NO CIGAR....");
	fprintf(fp,"\n\n<<<<<<<<<<<<<<<<<<<<<< END ROUTING PHASE >>>>>>>>>>>>>>>>>>>>>>>>\n\n");

      }

      fclose(fp);
      exit(1);

    }












    /* start measuring the anneal runtime */
    start_time=time(0);

    /* set up a random initial placement of the netlist on the FPGA */
    initPlacement();

/*      dLimit=(float)(FPGA->get_numRows()); */

    cout<<endl<<endl;
    cout<<endl<<"Netlist is "<<argv[1];
    cout<<endl<<"Total num blocks (LOGIC) = "<<numBlocks<<" and total num nets = "<<numNets;
    cout<<endl<<"Target is an HSRA with "<<FPGA->get_numCLBs()<<" logic blocks";
    cout<<endl<<"Target LSIZE is "<<FPGA->lsize;
    cout<<endl<<"The base channel width is "<<FPGA->get_baseChanWidth()<<" and growth rate is "<<FPGA->get_growthRate();
    cout<<endl;

    fprintf(fp,"\n\n\nNetlist is %s",argv[1]);
    fprintf(fp,"\nTotal num blocks (LOGIC) is %d, and total num nets is %d",numBlocks,numNets);
    fprintf(fp,"\nTarget is an HSRA with %d logic blocks",FPGA->get_numCLBs());
    fprintf(fp,"\nTarget LSIZE is ",FPGA->lsize);
    fprintf(fp,"\nThe base channel width is %d, and growth rate is %f",FPGA->get_baseChanWidth(),FPGA->get_growthRate());
    fprintf(fp,"\n");

#if defined(PLACER_DEBUG)
    printPlacement();
#endif

    /* to facilitate A*, sort sinks in increasing order of separation from source */
    netlist.sort_netSinks();

    /* Initial Pass...route each net in the netlist and calculate sharing and wire costs */
    routeNets(1,true);

    /* calculate starting temperature */
    calcStartTemperature();

    /* enter simulated annealing placement routine */
    annealDesign(fp);

    /* stop measuring the anneal runtime */
    end_time=time(0);

    cout<<endl<<endl<<"TOTAL ANNEAL RUNTIME IN SECONDS : "<<end_time-start_time;
    fprintf(fp,"\n\nTOTAL ANNEAL RUNTIME IN SECONDS : %d",end_time-start_time);

    fprintf(fp,"\n\nPLACEMENT IN HSRA FORMAT\n");
    printHSRAPlacement(fp);

/*      printPlacement(); */

/*      exit(1); */

    /* start measuring the routing runtime */
    start_time=time(0);

    ALPHA=0.0;

    /* rip up all nets */
/*      cout<<endl<<endl<<"RIPPING UP THE ENTIRE ROUTING...."; */
    netlist.create_netListItr();
    for(Net* net=netlist.getnext_netList();net!=0;net=netlist.getnext_netList()){
      if(net->sinkVector!=0)
	net->rip_rTree();
    }
    netlist.remove_netListItr();
/*      cout<<"DONE!! WIRE USAGE = "<<wireCost<<" AND SHARING = "<<sharingCost;  */

    /* do a final routing phase consisting of 50 iterations */
    FPGA->updateHistoryCosts(1,false);
    routeNets(1,true);
    sharingCost=FPGA->countSharing(0);wireCost=calcWireCost();
    cout<<endl<<endl<<"<<<<<<<<<<<<<<<<<<<<<< BEGIN ROUTING PHASE >>>>>>>>>>>>>>>>>>>>>>>>";
    fprintf(fp,"\n\n<<<<<<<<<<<<<<<<<<<<<< BEGIN ROUTING PHASE >>>>>>>>>>>>>>>>>>>>>>>>");
    cout<<endl<<"ITERATION 1\tSHARING = "<<sharingCost;
    fprintf(fp,"\nITERATION 1\tSHARING = %d",sharingCost);
    for(itr=2;itr<=200;itr++){
      FPGA->updateHistoryCosts(itr,false);
      routeNets(itr,true);
      cout<<endl<<"ITERATION "<<itr<<"\tSHARING = "<<sharingCost;
      fprintf(fp,"\nITERATION %d\tSHARING = %d",itr,sharingCost);
      if(sharingCost==0){
        cout<<endl<<endl<<"SUCCESFULLY PLACED AND ROUTED THIS BLOKE!!";
        cout<<endl<<endl<<"<<<<<<<<<<<<<<<<<<<<<< END ROUTING PHASE >>>>>>>>>>>>>>>>>>>>>>>>";
        cout<<endl<<endl;

	fprintf(fp,"\n\nSUCCESSFULLY PLACED AND ROUTED THIS BLOKE!!");
	fprintf(fp,"\n\n<<<<<<<<<<<<<<<<<<<<<< END ROUTING PHASE >>>>>>>>>>>>>>>>>>>>>>>>\n\n");

        break;
      }
    }

    if(itr==201){
      printNetsAndRecalcCosts();
      printNetsToFile(fp);
      cout<<endl<<endl<<"CLOSE, BUT NO CIGAR....";
      cout<<endl<<endl<<"<<<<<<<<<<<<<<<<<<<<<< END ROUTING PHASE >>>>>>>>>>>>>>>>>>>>>>>>";
      cout<<endl<<endl;

      fprintf(fp,"\n\nCLOSE, BUT NO CIGAR....");
      fprintf(fp,"\n\n<<<<<<<<<<<<<<<<<<<<<< END ROUTING PHASE >>>>>>>>>>>>>>>>>>>>>>>>\n\n");

    }

    /* start measuring the anneal runtime */
    end_time=time(0);

    cout<<endl<<endl<<"TOTAL ROUTING RUNTIME IN SECONDS : "<<end_time-start_time;
    cout<<endl<<endl;
    cout<<endl;

    fprintf(fp,"\n\nTOTAL ROUTING RUNTIME IN SECONDS : %d\n\n\n",end_time-start_time);

    fclose(fp);

    return 0;
  }
  catch(out_of_range){
    cerr<<"vector overflow error...\n";
  }

}


void slogEstimates(){
  time_t start_time,end_time;
  double est;

  start_time = time(NULL);
  for(int slog=0; slog<25 ; slog++){
    for(int i=0 ; i<FPGA->wire_rnodeVec->size() ; i++){
      for(int j=0 ; j<FPGA->sink_rnodeVec->size() ; j++)
	est = cost_to_target(FPGA->wire_rnodeVec->at(i),FPGA->sink_rnodeVec->at(j));
    }
  }
  end_time = time(NULL);

  cout<<endl<<endl<<"Total SLOGGG time is "<<(end_time - start_time);
  cout<<endl<<endl;
/*   exit(1); */

}


void populateAllSinksTables(){
  routingNode* rnode;
  priority_queue<pqNode>* PQ;

  FPGA->create_rnodeListItr();
  for(rnode=FPGA->getnext_rnodeList();rnode!=0;rnode=FPGA->getnext_rnodeList()){
    if(rnode->type==SOURCE){
      (rnode->allSinksTable).resize(FPGA->numSinkNodes);
      for(int i=0;i<(rnode->allSinksTable).size();i++)
	(rnode->allSinksTable).at(i)=-1.0;
      PQ=new priority_queue<pqNode>;
      PQ->push(pqNode(rnode,0.0,0.0,0));
      Dijkstra(PQ,rnode,0);
      delete PQ;
    }
  }
  FPGA->remove_rnodeListItr();
}

void dumpAllSinksTables(){
  routingNode* rnode;

  cout<<endl<<endl<<"<<<<<<<<< Begin Print Sink Distances >>>>>>>>>"<<endl; 

  FPGA->create_rnodeListItr();
  for(rnode=FPGA->getnext_rnodeList();rnode!=0;rnode=FPGA->getnext_rnodeList()){
    if(rnode->type==SOURCE){
      cout<<endl<<rnode->name<<":  ";
      for(int i=0;i<(rnode->allSinksTable.size());i++)
	cout<<(rnode->allSinksTable).at(i)<<"  ";
      cout<<endl;
    }
  }
  FPGA->remove_rnodeListItr();

  cout<<endl<<"<<<<<<<<< End Print Sink Distances >>>>>>>>>"<<endl; 

}


void dumpCostToTargetEstimates(){
  routingNode* rnode;routingNode* sink_rnode;
  list<routingNode*>::iterator r_itr;
  int estimate,rnode_cluster;

  FPGA->create_rnodeListItr();
  rnode=FPGA->getnext_rnodeList();
  for(;rnode!=0;rnode=FPGA->getnext_rnodeList()){
    if(rnode->type==BUS_OR_WIRE){
      cout<<endl<<endl<<"Routing Node "<<rnode->name<<"."; 
      cout<<endl;

      r_itr=FPGA->rnodeList.begin();
      for(;r_itr!=FPGA->rnodeList.end();r_itr++){
	if((*r_itr)->type==SINK){
	  sink_rnode = *r_itr;
	  estimate = (int)cost_to_target(rnode,sink_rnode);
	  cout<<"("<<sink_rnode->name<<","<<estimate<<")  ";
	}
      }
    }
  }
  FPGA->remove_rnodeListItr();

}


void performAstarSanityCheck(FILE* fp){
  int idx;int rnode_clus = -1;int sink_id;int actual_num_sinks = 0;
  routingNode* rnode;routingNode* sink_rnode;
  priority_queue<pqNode>* PQ;
  float running_sum = 0.0;float avg_error = 0.0;
  float true_cost=0.0;float estimated_cost=0.0;
  int underestimate_ctr=0;float avg_under=0.;

  cout<<endl<<endl<<"Performing Astar sanity check....";
  fprintf(fp,"\n\nPerforming Astar sanity check....");

  int total_num_wires = FPGA->wire_rnodeVec->size();
  int num_wires = (int)(ASTAR_SANITY_SAMPLE*(float)(total_num_wires));
  if(num_wires > total_num_wires)
    num_wires = total_num_wires;

  vector<bool> done (total_num_wires);
  for(int i=0;i<done.size();i++)
    done.at(i)=false;

  int num_sinks = FPGA->sink_rnodeVec->size();
  vector<float> cost_vec (num_sinks);
  for(int i=0;i<cost_vec.size();i++)
    cost_vec.at(i) = -1.0;

  for(int i=0;i<num_wires;i++){
    while(1){
      idx=lrand48()%total_num_wires;
      if(done.at(idx) || FPGA->wire_rnodeVec->at(idx)->canUse==false)
	continue;
      else
	break;
    }

    done.at(idx) = true;
    rnode = FPGA->wire_rnodeVec->at(idx);

    PQ=new priority_queue<pqNode>;
    PQ->push(pqNode(rnode,0.0,0.0,0));
    Dijkstra(PQ,rnode,&cost_vec);
    delete PQ;

    rnode_clus = rnode->clusterNumber;
/*     running_sum=0.;actual_num_sinks=0;underestimate_ctr=0; */
    for(int j=0;j<num_sinks;j++){
      sink_rnode = FPGA->sink_rnodeVec->at(j);
      sink_id = sink_rnode->sinkID;
      true_cost = cost_vec.at(sink_id);
      estimated_cost = cost_to_target(rnode,sink_rnode);

      if(true_cost < 0.0)  // unreachable sinks have -ve cost
	continue;

      actual_num_sinks++;
      if(estimated_cost > true_cost){
	cout<<endl<<endl;
	cout<<endl<<"Wire is "<<rnode->name<<". Sink is "<<sink_rnode->name<<".";
	fprintf(fp,"\n\n\nWire is %s. Sink is %s.",rnode->name,sink_rnode->name);
	cout<<endl<<"Estimated cost is "<<estimated_cost<<", and true cost is "<<true_cost<<".";
	fprintf(fp,"\nEstimated cost is %4.2f, and true cost is %4.2f.",estimated_cost,true_cost); 
	cout<<endl<<"A* estimate exceeds actual path cost here!!";
	cout<<endl<<endl;
	fprintf(fp,"\nA* estimate exceeds actual path cost here!!\n\n");
	fclose(fp);
	exit(1);
      }

      running_sum+=(true_cost - estimated_cost)/true_cost;
      if(estimated_cost < true_cost-0.01)
	underestimate_ctr++;

    }

    for(int y=0;y<cost_vec.size();y++)
      cost_vec.at(y) = -1.0;

  }

  avg_error = running_sum/(float)underestimate_ctr;
  avg_under = (float)underestimate_ctr/(float)actual_num_sinks;

  cout<<"done!!";
  fprintf(fp,"done!!");
/*   cout<<endl<<"Total num sinks is "<<actual_num_sinks; */
/*   fprintf(fp,"done!!\nTotal num sinks is %d",actual_num_sinks); */
  cout<<endl<<"Total num wires is "<<total_num_wires<<", and num wires used to perform sanity check is "<<num_wires;
  fprintf(fp,"\nTotal num wires is %d, and num wires used to perform sanity check is %d",total_num_wires,num_wires);
  cout<<endl<<"Average cost underestimate is "<<avg_error<<".";
  fprintf(fp,"\nAverage cost underestimate is %f.",avg_error);
  cout<<endl<<"Average fraction sinks underestimated is "<<avg_under<<"."<<endl;
  fprintf(fp,"\nAverage fraction sinks underestimated is %f.",avg_under);
  fprintf(fp,"\n");

}

void Dijkstra(priority_queue<pqNode>* PQ,routingNode* source_rnode,vector<float>* cost_vec){
  routingNode* rnode;routingNode* nb_rnode;
  routingEdge* nb_edge;
  double path_cost,nb_path_cost,new_path_cost;
  list<routingNode*> cleanupList;
  list<routingNode*>::iterator clean_itr;

/*   cout<<endl<<endl<<"Looking at source node "<<source_rnode->name<<", index = "<<source_rnode->nodeIndex<<"."<<endl; */

  while(!PQ->empty()){

    rnode=(PQ->top()).rNode;
    if(rnode->isExplored==false){
      /* found unexplored routing node...mark it explored,set its pathcost and */
      /* predecessor routing node and break */
      rnode->isExplored=true;
      rnode->pathCost=(PQ->top()).pathCost;
      rnode->predecessor=(PQ->top()).predNode;

      /* fill in cost to sink rnode from source_rnode in costTable */
      if(rnode->get_type()==SINK){
/* 	cout<<"("<<rnode->name<<",index="<<rnode->sinkIndex<<",pathCost="<<rnode->pathCost<<")  "; */
	if(cost_vec==0)
	  (source_rnode->allSinksTable).at(rnode->sinkID)=rnode->pathCost;
	else
	  cost_vec->at(rnode->sinkID)=rnode->pathCost;
      }

      cleanupList.push_back(rnode);  //this list holds routing nodes that need a clean-up
      PQ->pop();
    }
    else{
      PQ->pop();
      continue;
    }
 
    /* loop thru the fanout neighbor nodes of rnode */ 
    rnode->create_fanoutListItr();
    for(nb_edge=rnode->getnext_fanoutList();nb_edge!=0;nb_edge=rnode->getnext_fanoutList()){
      nb_rnode=nb_edge->destNode;
      if(nb_rnode->isExplored==false){
        path_cost=rnode->pathCost;
	new_path_cost=path_cost+nb_rnode->returnCost(1,true);
	nb_path_cost=nb_rnode->pathCost;
	if(new_path_cost<nb_path_cost)
	  PQ->push(pqNode(nb_rnode,new_path_cost,0.,rnode));
      }
    }
    rnode->remove_fanoutListItr();

  }
  /* end exploration */

  /* clean up all routing nodes that have been affected by the exploration */
  for(clean_itr=cleanupList.begin();clean_itr!=cleanupList.end();clean_itr++){
    (*clean_itr)->isExplored=false;
    (*clean_itr)->inrTree=false;
    (*clean_itr)->predecessor=0;
    (*clean_itr)->pathCost=HUGE_NUMBER;
    (*clean_itr)->isRouted=false;
  }

}


void dumpAstarTable(){

  cout<<endl<<endl<<"<<<<<<<<<<<< BEGIN DUMP A* TABLE >>>>>>>>>>>>"<<endl;

  cout<<endl<<"NUM_CLUSTERS  "<<AstarTable->size();
  cout<<endl<<"NUM_SINKS  "<<AstarTable->at(0)->size();

  for(int i=0;i<AstarTable->size();i++){
    cout<<endl<<endl<<"CLUSTER NUM: "<<i<<endl;
    for(int j=0;j<AstarTable->at(i)->size();j++)
      cout<<AstarTable->at(i)->at(j)<<"  ";
  }

  cout<<endl<<endl<<"<<<<<<<<<<<<< END DUMP A* TABLE >>>>>>>>>>>>>";

}

void process_AstarTableEntry(int cluster_id,char* sink_name,double estimate){

  if(sink_name==0 || cluster_id<0 || cluster_id>=numClusters)
    ERROR("error while processing Astar table entry...");

  string rname = sink_name;

  routingNode* rnode=FPGA->rnodeMap[rname];
  if(rnode==0){
    cout<<endl<<sink_name<<endl;
    ERROR("Could not find rnode in list of routing nodes...\n");
  }

/*   cout<<endl<<"Sink node is "<<rname<<", estimate is "<<estimate<<", cluster_id = "<<cluster_id<<", sinkID = "<<rnode->sinkID; */

  AstarTable->at(cluster_id)->at(rnode->sinkID)=estimate;
}

void prepareAstarTable(int num_clusters,int num_sinks){
  string rname;

  if(num_clusters<1 || num_sinks<1)
    ERROR("dubious Astar information here...");

  /* allocate table of Astar estimates */
  AstarTable=new vector<vector<double>*> (num_clusters);
  for(int i=0;i<AstarTable->size();i++){
    AstarTable->at(i)=new vector<double> (num_sinks);
    for(int j=0;j<AstarTable->at(i)->size();j++)
      AstarTable->at(i)->at(j)=-1.0;
  }

}

void readInClusterAssignments(char* clus_fname){
  char rname [MAX_STR_LEN];
  int cluster_num=-1;
  routingNode* rnode;string rstring;

  ifstream clus_fstream(clus_fname);
  if(!clus_fstream)
    ERROR("could not open cluster assignments file...");

  while(!clus_fstream.eof()){
    clus_fstream >> rname;
    clus_fstream >> cluster_num;
    rstring = rname;
    rnode = FPGA->rnodeMap[rstring];
    if(rnode==0){
      cout<<endl<<rname<<endl;
      ERROR("could not find rnode..."); 
    }
    rnode->clusterNumber = cluster_num;
  }
  clus_fstream.close();

}

void dumpClusterAssignments(){
  routingNode* rnode;

  FPGA->create_rnodeListItr();
  for(rnode=FPGA->getnext_rnodeList();rnode!=0;rnode=FPGA->getnext_rnodeList())
    cout<<endl<<rnode->name<<"\t"<<rnode->clusterNumber;
  FPGA->remove_rnodeListItr();

}


void dumpArchInRRGraph(FILE* fp){
  routingNode* rnode;

  list<routingNode*>::iterator itr = (FPGA->rnodeList).begin();
  for(;itr!=(FPGA->rnodeList).end();itr++){
    rnode=*itr;
    if(rnode->type==SOURCE)
      writeOutSourceNode(rnode,fp);
    else if(rnode->type==SINK)
      writeOutSinkNode(rnode,fp);
    else if(rnode->type==BUS_OR_WIRE)
      writeOutWireNode(rnode,fp);
    else if(rnode->type==INPUT_PIN)
      writeOutInputPinNode(rnode,fp);
    else if(rnode->type==OUTPUT_PIN)
      writeOutOutputPinNode(rnode,fp);
  }

  /* finally dump the IOB source and sink nodes */
/*   writeOutSourceNode(FPGA->iobSource,fp); */
/*   writeOutSinkNode(FPGA->iobSink,fp); */

}

void writeOutSourceNode(routingNode* rnode,FILE* fp){
  int edge_type;routingNode* dest_rnode;

  fprintf(fp,"Node: %s.  SOURCE (-1, -1)  Ptc_num: -1",rnode->name);

  fprintf(fp,"\n   %d edge(s):",(rnode->fanoutList).size());
  list<routingEdge*>::iterator redge_itr = (rnode->fanoutList).begin();
  for(;redge_itr!=(rnode->fanoutList).end();redge_itr++){
    dest_rnode=(*redge_itr)->destNode;
    fprintf(fp," %s",dest_rnode->name);
  }

  fprintf(fp,"\nSwitch types:");
  redge_itr = (rnode->fanoutList).begin();
  for(;redge_itr!=(rnode->fanoutList).end();redge_itr++){
    edge_type=(*redge_itr)->type;
    fprintf(fp," %d",edge_type);
  }

  fprintf(fp,"\nOcc: 0  Capacity: %d.",rnode->capacity);
  fprintf(fp,"\nR: 0  C: 0");
  fprintf(fp,"\nCost_index: 0\n\n");
}

void writeOutOutputPinNode(routingNode* rnode,FILE* fp){
  int edge_type;routingNode* dest_rnode;

  fprintf(fp,"Node: %s.  OPIN (-1, -1)  Ptc_num: -1",rnode->name);

  fprintf(fp,"\n   %d edge(s):",(rnode->fanoutList).size());
  list<routingEdge*>::iterator redge_itr = (rnode->fanoutList).begin();
  for(;redge_itr!=(rnode->fanoutList).end();redge_itr++){
    dest_rnode=(*redge_itr)->destNode;
    fprintf(fp," %s",dest_rnode->name);
  }

  fprintf(fp,"\nSwitch types:");
  redge_itr = (rnode->fanoutList).begin();
  for(;redge_itr!=(rnode->fanoutList).end();redge_itr++){
    edge_type=(*redge_itr)->type;
    fprintf(fp," %d",edge_type);
  }

  fprintf(fp,"\nOcc: 0  Capacity: 1.");
  fprintf(fp,"\nR: 0  C: 0");
  fprintf(fp,"\nCost_index: 0\n\n");
}

void writeOutSinkNode(routingNode* rnode,FILE* fp){
  int edge_type;

  fprintf(fp,"Node: %s.  SINK (-1, -1)  Ptc_num: -1",rnode->name);
  fprintf(fp,"\n   0 edge(s):");
  fprintf(fp,"\nSwitch types:");
  fprintf(fp,"\nOcc: 0  Capacity: %d.",rnode->capacity);
  fprintf(fp,"\nR: 0  C: 0");
  fprintf(fp,"\nCost_index: 0\n\n");
}

void writeOutInputPinNode(routingNode* rnode,FILE* fp){
  int edge_type;routingNode* dest_rnode;

  fprintf(fp,"Node: %s.  IPIN (-1, -1)  Ptc_num: -1",rnode->name);

  fprintf(fp,"\n   %d edge(s):",(rnode->fanoutList).size());
  list<routingEdge*>::iterator redge_itr = (rnode->fanoutList).begin();
  for(;redge_itr!=(rnode->fanoutList).end();redge_itr++){
    dest_rnode=(*redge_itr)->destNode;
    fprintf(fp," %s",dest_rnode->name);
  }

  fprintf(fp,"\nSwitch types:");
  redge_itr = (rnode->fanoutList).begin();
  for(;redge_itr!=(rnode->fanoutList).end();redge_itr++){
    edge_type=(*redge_itr)->type;
    fprintf(fp," %d",edge_type);
  }

  fprintf(fp,"\nOcc: 0  Capacity: 1.");
  fprintf(fp,"\nR: 0  C: 0");
  fprintf(fp,"\nCost_index: 0\n\n");
}

void writeOutWireNode(routingNode* rnode,FILE* fp){
  int edge_type;routingNode* dest_rnode;

  fprintf(fp,"Node: %s.  CHANX (-1, -1)  Ptc_num: -1",rnode->name);

  fprintf(fp,"\n   %d edge(s):",(rnode->fanoutList).size());
  list<routingEdge*>::iterator redge_itr = (rnode->fanoutList).begin();
  for(;redge_itr!=(rnode->fanoutList).end();redge_itr++){
    dest_rnode=(*redge_itr)->destNode;
    fprintf(fp," %s",dest_rnode->name);
  }

  fprintf(fp,"\nSwitch types:");
  redge_itr = (rnode->fanoutList).begin();
  for(;redge_itr!=(rnode->fanoutList).end();redge_itr++){
    edge_type=(*redge_itr)->type;
    fprintf(fp," %d",edge_type);
  }

  fprintf(fp,"\nOcc: 0  Capacity: 1.");
  fprintf(fp,"\nR: 0  C: 0");
  fprintf(fp,"\nCost_index: 0\n\n");
}

void calcStartTemperature(){
  hsraBlock* block0;hsraBlock* block1;
  vector<float> costArray(numBlocks);
  float prev_wireCost=(float)(wireCost);
  int index0,index1,block_type;

  for(int i=0;i<numBlocks;i++){

    /* randomly generate a pair of FPGA tiles that will be involved in the move */
    generateMove(&block0,&block1,&block_type);

    /* make the move i.e. actually move the logic/IO blocks from one location to the other */
    makeMove(&block0,&block1,&block_type,1);

#if defined(PLACER_DEBUG)
    actual_wireCost=calcWireCost();
    actual_sharingCost=FPGA->countSharing(0);
    if(wireCost!=actual_wireCost)
      ERROR("wire calculation mismatch...");
    if(sharingCost!=actual_sharingCost)
      ERROR("sharing calculation mismatch...");
#endif

    /* sort sinks for A* purposes */
    sortSinks(&block0,&block1,&block_type);

    costArray.at(i)=SHARE_FAC*(float)(sharingCost) + WIRE_FAC*(float)(wireCost);

  }

  float avg_cost=0.;float cov=0.;
  /* calculate the average cost */
  for(int j=0;j<numBlocks;j++)
    avg_cost+=costArray.at(j)/(float)(numBlocks);
  /* calculate cov */
  for(int k=0;k<numBlocks;k++)
    cov+=(costArray.at(k) - avg_cost)*(costArray.at(k) - avg_cost);

  annealTemp = (float)(0.1 * pow((double)cov/(double)(numBlocks),0.5));

}

bool thereIsANeedToAnneal(){

  if(annealTemp<1.0e-5)
    return false;
  else
    return true;

}

void annealDesign(FILE* fp){
  int bf_sharingCost,bf_wireCost,af_sharingCost,af_wireCost,prev_wireCost;
  int iter=1;float deltaC;double i;
  hsraBlock* block0;hsraBlock* block1;
  int block_type;
  list<cachedObj*> sink_list0;list<cachedObj*> sink_list1;
  list<cachedObj*> src_list0;list<cachedObj*> src_list1;
  double num_acc=0.;double fracn_acc=1.;double alpha=0.;
  float placementCost=0.;
  time_t startTime,stopTime;

  cout<<endl<<endl;
  cout<<endl<<"temp.\t\tw_use\t\tshare\t\taccept\t\tdlim\t    cost\tALPHA";
  cout<<endl<<"===================================================================================================";

  fprintf(fp,"\n\n\ntemp.\t\tw_use\t\tshare\t\taccept\t\tdlim\t    cost\tALPHA");
  fprintf(fp,"\n===================================================================================================");

/*    while(annealTemp >= (0.0005*(SHARE_FAC*(float)(sharingCost)+(float)(wireCost))) / (float)(numNets) ){ */
  while(fracn_acc > 0.03){
    iter++;

    FPGA->updateHistoryCosts(iter,true);

    prev_wireCost=wireCost;

    num_acc=0.;

    for(i=0.;i<5.0*pow(numBlocks,1.33);i+=1.0){

#if defined(PLACER_DEBUG)
      cout<<endl<<endl<<"********************** GENERATING NEW MOVE ************************";
#endif


      generateMove(&block0,&block1,&block_type);


      bf_sharingCost=sharingCost;bf_wireCost=wireCost;


      cacheRoutes(&block0,&block1,&block_type,&src_list0,&src_list1,&sink_list0,&sink_list1);

 
      makeMove(&block0,&block1,&block_type,iter);


      af_sharingCost=sharingCost;af_wireCost=wireCost;

      deltaC=SHARE_FAC*(float)(af_sharingCost-bf_sharingCost)/(float)(prev_wireCost) + 
             WIRE_FAC*(float)(af_wireCost-bf_wireCost)/(float)(prev_wireCost);

      if(deltaC > 1.0e-10){
	// this is a bad move

        if(drand48() < exp(-0.1*deltaC/annealTemp)){
          // accept this move
	  sortSinks(&block0,&block1,&block_type);
          clearCachedRoutes(&src_list0,&src_list1,&sink_list0,&sink_list1,1);
          num_acc+=1.0;

#if defined(PLACER_DEBUG)
          actual_wireCost=calcWireCost();
          actual_sharingCost=FPGA->countSharing(0);
          if(wireCost!=actual_wireCost)
            ERROR("wire calculation mismatch...");
          if(sharingCost!=actual_sharingCost)
            ERROR("sharing calculation mismatch...");
#endif

        }
        else{
          // reject this move
          undoMove(&block0,&block1,&block_type,&src_list0,&src_list1,&sink_list0,&sink_list1,iter);
          clearCachedRoutes(&src_list0,&src_list1,&sink_list0,&sink_list1,0);

#if defined(PLACER_DEBUG)
          actual_wireCost=calcWireCost();
          actual_sharingCost=FPGA->countSharing(0);
          if(wireCost!=actual_wireCost)
            ERROR("wire calculation mismatch...");
          if(sharingCost!=actual_sharingCost)
            ERROR("sharing calculation mismatch...");
#endif

        }

      }
      else{
	// this is a good move...keep it
	sortSinks(&block0,&block1,&block_type);
	clearCachedRoutes(&src_list0,&src_list1,&sink_list0,&sink_list1,1);

	if(deltaC < -1.0e-10)
	  num_acc+=1.0;
/*  	num_acc+=1.0; */

#if defined(PLACER_DEBUG)
        actual_wireCost=calcWireCost();
        actual_sharingCost=FPGA->countSharing(0);
        if(wireCost!=actual_wireCost)
          ERROR("wire calculation mismatch...");
        if(sharingCost!=actual_sharingCost)
          ERROR("sharing calculation mismatch...");
#endif

      }

    }

    stopTime=time(0);

    routeNets(iter,true);

    fracn_acc=num_acc/i;
    placementCost=SHARE_FAC*(float)(sharingCost) + WIRE_FAC*(float)(wireCost);
    printf("\n%9.7f\t %ld\t\t  %ld\t\t%5.4f\t\t%4.1f\t%10.2f\t%4.2f",annealTemp,wireCost,sharingCost,fracn_acc,dLimit,placementCost,ALPHA);
    fprintf(fp,"\n%9.7f\t %ld\t\t  %ld\t\t%5.4f\t\t%4.1f\t%10.2f\t%4.2f",annealTemp,wireCost,sharingCost,fracn_acc,dLimit,placementCost,ALPHA);

/*      if(sharingCost==0) */
/*        return; */

    if(fracn_acc>0.96)
      alpha=0.5;
    else if(fracn_acc<=0.96 && fracn_acc>0.8)
      alpha=0.9;
    else if(fracn_acc<=0.8 && fracn_acc>0.15)
      alpha=0.95;
    else
      alpha=0.8;

/*      if(fracn_acc>0.85) */
/*        alpha=0.5; */
/*      else if(fracn_acc<=0.85 && fracn_acc>0.50) */
/*        alpha=0.95; */
/*      else if(fracn_acc<=0.50 && fracn_acc>0.03) */
/*        alpha=0.99; */

    annealTemp=alpha*annealTemp;

  }

  /* quench now */
  for(int n=0;n<0;n++){

    /* increment iteration */
    iter++;
    /* update history cost */
    FPGA->updateHistoryCosts(iter,true);
    /* init normalizing factor */
    prev_wireCost=wireCost;
    /* init num moves accepted to 0 */
    num_acc=0.;

    for(i=0.;i<5.0*pow(numBlocks,1.33);i+=1.0){

#if defined(PLACER_DEBUG)
      cout<<endl<<endl<<"********************** GENERATING NEW MOVE ************************";
#endif

      /* randomly generate a pair of FPGA tiles that will be involved in the move */
      generateMove(&block0,&block1,&block_type);

      /* cache sharing and wire costs before making the move */
      bf_sharingCost=sharingCost;bf_wireCost=wireCost;

      /* cache route trees & terminal mapping of the nets that connect to the moved blocks */
      cacheRoutes(&block0,&block1,&block_type,&src_list0,&src_list1,&sink_list0,&sink_list1);

      /* make the move i.e. actually move the logic/IO blocks from one location to the other */
      makeMove(&block0,&block1,&block_type,iter);

      /* cache sharing and wire costs after making the move */
      af_sharingCost=sharingCost;af_wireCost=wireCost;

      deltaC=SHARE_FAC*(float)(af_sharingCost-bf_sharingCost)/(float)(prev_wireCost) + 
             WIRE_FAC*(float)(af_wireCost-bf_wireCost)/(float)(prev_wireCost);

      if(deltaC > 1.0e-10){
	/* this is a bad move....reject */
	undoMove(&block0,&block1,&block_type,&src_list0,&src_list1,&sink_list0,&sink_list1,iter);
        clearCachedRoutes(&src_list0,&src_list1,&sink_list0,&sink_list1,0);

#if defined(PLACER_DEBUG)
        actual_wireCost=calcWireCost();
        actual_sharingCost=FPGA->countSharing(0);
        if(wireCost!=actual_wireCost)
          ERROR("wire calculation mismatch...");
        if(sharingCost!=actual_sharingCost)
          ERROR("sharing calculation mismatch...");
#endif

      }
      else{
	/* this is a good move...keep it */
	sortSinks(&block0,&block1,&block_type);
	clearCachedRoutes(&src_list0,&src_list1,&sink_list0,&sink_list1,1);

	if(deltaC < -1.0e-10)
	  num_acc+=1.0;
/*  	num_acc+=1.0; */

#if defined(PLACER_DEBUG)
        actual_wireCost=calcWireCost();
        actual_sharingCost=FPGA->countSharing(0);
        if(wireCost!=actual_wireCost)
          ERROR("wire calculation mismatch...");
        if(sharingCost!=actual_sharingCost)
          ERROR("sharing calculation mismatch...");
#endif

      }

    }

    fracn_acc=num_acc/i;
    placementCost=SHARE_FAC*(float)(sharingCost) + WIRE_FAC*(float)(wireCost);
    printf("\n%9.7f\t %ld\t\t  %ld\t\t%5.4f\t\t%4.1f\t\t%10.2f",annealTemp,wireCost,sharingCost,fracn_acc,dLimit,placementCost);
    fprintf(fp,"\n%9.7f\t %ld\t\t  %ld\t\t%5.4f\t\t%4.1f\t\t%10.2f",annealTemp,wireCost,sharingCost,fracn_acc,dLimit,placementCost);

/*      routeNets(iter,true); */

  }

}


void undoMove(hsraBlock** block0,hsraBlock** block1,int* block_type,list<cachedObj*>* src_list0,
              list<cachedObj*>* src_list1,list<cachedObj*>* sink_list0,
              list<cachedObj*>* sink_list1,int iter){
  logicBlock* lb0;logicBlock* lb1;
  Terminal* term;
  int connected=0;

  lb0=(*block0)->get_currBlock();
  lb1=(*block1)->get_currBlock();

  /* alright, rip up the source nets and sink segments that connect to lb0 and/or lb1 */
  ripRoutes(lb0,lb1);

  /* clear the terminal lists of the SOURCE and SINK routing nodes in block0 and block1 */
  (*block0)->get_sourceNode()->clear_terminalList();
  (*block0)->get_sinkNode()->clear_terminalList();
  (*block1)->get_sourceNode()->clear_terminalList();
  (*block1)->get_sinkNode()->clear_terminalList();

  /* translate/swap lb0 and/or lb1 */
  placeBlock(lb1,*block0,*block_type);
  if(lb0)
    placeBlock(lb0,*block1,*block_type);
  else
    (*block1)->set_currBlock(0);

  lb0=(*block0)->get_currBlock();
  lb1=(*block1)->get_currBlock();

  /* now reconnect the previously cached nets and net segments to the terminals of lb0 */
  /* and/or lb1. Route trees that were cached in cacheRoutes() will be used here. */

  if(lb0){
    /* loop thru the output terminals of lb0. for each output terminal, reconnect the */
    /* cached net back to lb0. */
    list<cachedObj*>::iterator list_itr=src_list0->begin();
    lb0->create_outputListItr();
    for(term=lb0->getnext_outputList();term!=0;term=lb0->getnext_outputList()){
      if(term->get_net()->get_isCounted()==false){
        /* only consider those nets that have not already been ripped up */

	if((*list_itr)->rNode!=term->get_rNode())
	  ERROR("terminal mismatch while undoing move...");
	/* repopulate the net's route tree with previously cached route tree. */
	(term->get_net()->get_rTree())->set_root(((*list_itr)->rTree).get_root());
	/* increment occupancies */
	(term->get_net()->get_rTree())->incrementOccupancies(0);
	term->get_net()->set_isCounted(true);
	/* increment the source list iterator */
	list_itr++;
 
      }
    }
    lb0->remove_outputListItr();
  }

  if(lb1){
    /* loop thru the output terminals of lb1. for each output terminal, reconnect the */
    /* cached net back to lb1. */
    list<cachedObj*>::iterator list_itr=src_list1->begin();
    lb1->create_outputListItr();
    for(term=lb1->getnext_outputList();term!=0;term=lb1->getnext_outputList()){
      if(term->get_net()->get_isCounted()==false){
        /* only consider those nets that have not already been ripped up */

	if((*list_itr)->rNode!=term->get_rNode())
	  ERROR("terminal mismatch while undoing move...");
	/* repopulate the net's route tree with previously cached route tree. */
	(term->get_net()->get_rTree())->set_root(((*list_itr)->rTree).get_root());
	/* increment occupancies. */
	(term->get_net()->get_rTree())->incrementOccupancies(0);
	term->get_net()->set_isCounted(true);
	/* increment the source list iterator */
	list_itr++;
 
      }
    }
    lb1->remove_outputListItr();
  }

  if(lb0){
    /* now, loop thru the input terminals of lb0. for each input terminal, reconnect */
    /* the previously cached net segment back to the sink terminal of the logic block */
    list<cachedObj*>::iterator list_itr=sink_list0->begin();
    lb0->create_inputListItr();
    for(term=lb0->getnext_inputList();term!=0;term=lb0->getnext_inputList()){
      if(term->get_net()->get_isCounted()==false){
        /* only consider those nets that have not already been ripped up */

	if((*list_itr)->rNode!=term->get_rNode())
	  ERROR("terminal mismatch while undoing move...");
	/* now reconnect the previously cached net segment to the sink terminal. */
	/* while doing so, increment occupancies. */
	connected=term->get_net()->reconnect(((*list_itr)->rTree).get_root());

        /* if unable to connect, store the terminal and branch. the connection will be made later. */
        if(connected==-1){
          routeSink(iter,term->get_net(),term->get_rNode(),1);
          (*list_itr)->clearTree=1;
        }

	/* increment the source list iterator */
	list_itr++;

      }
    }
    lb0->remove_inputListItr();
  }

  if(lb1){
    /* now, loop thru the input terminals of lb1. for each input terminal, reconnect */
    /* the previously cached net segment back to the sink terminal of the logic block */
    list<cachedObj*>::iterator list_itr=sink_list1->begin();
    lb1->create_inputListItr();
    for(term=lb1->getnext_inputList();term!=0;term=lb1->getnext_inputList()){
      if(term->get_net()->get_isCounted()==false){
        /* only consider those nets that have not already been ripped up */

	if((*list_itr)->rNode!=term->get_rNode())
	  ERROR("terminal mismatch while undoing move...");
	/* now reconnect the previously cached net segment to the sink terminal. */
	/* while doing so, increment occupancies. */
	connected=term->get_net()->reconnect(((*list_itr)->rTree).get_root());

        /* if unable to connect, store the terminal and branch. the connection will be made later. */
        if(connected==-1){
          routeSink(iter,term->get_net(),term->get_rNode(),1);
          (*list_itr)->clearTree=1;
        }

	/* increment the source list iterator */
	list_itr++;

      }
    }
    lb1->remove_inputListItr();
  }

  /* reset the output nets of lb0 and lb1 */
  if(lb0){
    lb0->create_outputListItr();
    for(term=lb0->getnext_outputList();term!=0;term=lb0->getnext_outputList())
      term->get_net()->set_isCounted(false);
    lb0->remove_outputListItr();
  }
  if(lb1){
    lb1->create_outputListItr();
    for(term=lb1->getnext_outputList();term!=0;term=lb1->getnext_outputList())
      term->get_net()->set_isCounted(false);
    lb1->remove_outputListItr();
  }

}

void clearCachedRoutes(list<cachedObj*>* src_list0,list<cachedObj*>* src_list1,
                list<cachedObj*>* sink_list0,list<cachedObj*>* sink_list1,int clear_tree){
  cachedObj* ptr;

  list<cachedObj*>::iterator list_itr=src_list0->begin();
  for(;list_itr!=src_list0->end();list_itr++){
    if(clear_tree==1 || (clear_tree==0 && (*list_itr)->clearTree==1))
      ((*list_itr)->rTree).clear();
    ptr=*list_itr;
    delete ptr;
  }
  src_list0->clear();

  list_itr=src_list1->begin();
  for(;list_itr!=src_list1->end();list_itr++){
    if(clear_tree==1 || (clear_tree==0 && (*list_itr)->clearTree==1))
      ((*list_itr)->rTree).clear();
    ptr=*list_itr;
    delete ptr;
  }
  src_list1->clear();

  list_itr=sink_list0->begin();
  for(;list_itr!=sink_list0->end();list_itr++){
    if(clear_tree==1 || (clear_tree==0 && (*list_itr)->clearTree==1))
      ((*list_itr)->rTree).clear();
    ptr=*list_itr;
    delete ptr;
  }
  sink_list0->clear();

  list_itr=sink_list1->begin();
  for(;list_itr!=sink_list1->end();list_itr++){
    if(clear_tree==1 || (clear_tree==0 && (*list_itr)->clearTree==1))
      ((*list_itr)->rTree).clear();
    ptr=*list_itr;
    delete ptr;
  }
  sink_list1->clear();

}

void cacheRoutes(hsraBlock** block0,hsraBlock** block1,int* block_type,list<cachedObj*>* src_list0,
                 list<cachedObj*>* src_list1,list<cachedObj*>* sink_list0,
                 list<cachedObj*>* sink_list1){
  logicBlock* lb0;logicBlock* lb1;
  Terminal* term;
  cachedObj* cache_obj;

  lb0=(*block0)->get_currBlock();
  lb1=(*block1)->get_currBlock();

#if defined(PLACER_DEBUG)
  cout<<endl<<endl<<"IN CACHE ROUTES...";
#endif

  if(lb0){

#if defined(PLACER_DEBUG)
    cout<<endl<<"LOOKING AT "<<lb0->get_name();
#endif

    /* loop thru the output terminals of lb0. for each output terminal, store the routing */
    /* node that the terminal maps to and the ENTIRE route tree of the net that it drives */
    lb0->create_outputListItr();
    for(term=lb0->getnext_outputList();term!=0;term=lb0->getnext_outputList()){ 
      /* create new cache object */
      cache_obj=new cachedObj;
      /* store routing node that the terminal maps to */
      cache_obj->rNode=term->get_rNode();
      /* copy ENTIRE route tree */
      (cache_obj->rTree).populateWith(term->get_net()->get_rTree()->get_root());
      /* add to list of cached objects */
      src_list0->push_back(cache_obj);

#if defined(PLACER_DEBUG)
      (cache_obj->rTree).print_info();
#endif

      term->get_net()->set_isCounted(true);
    }
    lb0->remove_outputListItr();
  }

  if(lb1){

#if defined(PLACER_DEBUG)
    cout<<endl<<"LOOKING AT "<<lb1->get_name();
#endif

    /* loop thru the output terminals of lb1. for each output terminal, store the routing */
    /* node that the terminal maps to and the ENTIRE route tree of the net that it drives */
    lb1->create_outputListItr();
    for(term=lb1->getnext_outputList();term!=0;term=lb1->getnext_outputList()){
      /* create new cache object */
      cache_obj=new cachedObj;
      /* store routing node that the terminal maps to */
      cache_obj->rNode=term->get_rNode();
      /* copy ENTIRE route tree */
      (cache_obj->rTree).populateWith((term->get_net()->get_rTree())->get_root());
      /* add to list of cached objects */
      src_list1->push_back(cache_obj);

#if defined(PLACER_DEBUG)
      (cache_obj->rTree).print_info();
#endif

      term->get_net()->set_isCounted(true);
    }
    lb1->remove_outputListItr();
  }

  if(lb0){

#if defined(PLACER_DEBUG)
    cout<<endl<<"LOOKING AT "<<lb0->get_name();
#endif

    /* now, loop thru the input terminals of lb0. for each input terminal, store the */
    /* routing node that it maps to and the segment of the net's route tree that sinks */
    /* at this input terminal */
    lb0->create_inputListItr();
    for(term=lb0->getnext_inputList();term!=0;term=lb0->getnext_inputList()){
      if(term->get_net()->get_isCounted()==false){
        /* only consider those nets that have not already been ripped up */

	/* create new cache object */
	cache_obj=new cachedObj;
	/* store routing node that the terminal maps to */
	cache_obj->rNode=term->get_rNode();
	/* get the sub-tree that sinks at this terminal and store a copy */
	(cache_obj->rTree).populateWith(
         term->get_net()->get_rTree()->get_subTree(term->get_rNode()),term->get_rNode());

#if defined(PLACER_DEBUG)
	(cache_obj->rTree).print_info();
#endif

	/* add to list of cached objects */
	sink_list0->push_back(cache_obj);

      }
    }
    lb0->remove_inputListItr();
  }

  if(lb1){

#if defined(PLACER_DEBUG)
    cout<<endl<<"LOOKING AT "<<lb1->get_name();
#endif

    /* now, loop thru the input terminals of lb1. for each input terminal, store the */
    /* routing node that it maps to and the segment of the net's route tree that sinks */
    /* at this input terminal */
    lb1->create_inputListItr();
    for(term=lb1->getnext_inputList();term!=0;term=lb1->getnext_inputList()){
      if(term->get_net()->get_isCounted()==false){
        /* only consider those nets that have not already been ripped up */

	/* create new cache object */
	cache_obj=new cachedObj;
	/* store routing node that the terminal maps to */
	cache_obj->rNode=term->get_rNode();
	/* get the sub-tree that sinks at this terminal and store a copy */
	(cache_obj->rTree).populateWith(
         term->get_net()->get_rTree()->get_subTree(term->get_rNode()),term->get_rNode());

#if defined(PLACER_DEBUG)
	(cache_obj->rTree).print_info();
#endif

	/* add to list of cached objects */
	sink_list1->push_back(cache_obj);

      }
    }
    lb1->remove_inputListItr();
  }

  /* reset the output nets of lb0 and lb1 */
  if(lb0){
    lb0->create_outputListItr();
    for(term=lb0->getnext_outputList();term!=0;term=lb0->getnext_outputList())
      term->get_net()->set_isCounted(false);
    lb0->remove_outputListItr();
  }
  if(lb1){
    lb1->create_outputListItr();
    for(term=lb1->getnext_outputList();term!=0;term=lb1->getnext_outputList())
      term->get_net()->set_isCounted(false);
    lb1->remove_outputListItr();
  }

#if defined(PLACER_DEBUG)
  cout<<endl<<"EXITING CACHE ROUTES...";
#endif

}

void sortSinks(hsraBlock** block0,hsraBlock** block1,int* block_type){
  Terminal* term;
  logicBlock* lb0;logicBlock* lb1;

  lb0=(*block0)->get_currBlock();
  lb1=(*block1)->get_currBlock();

  if(lb0){
    lb0->create_outputListItr();
    for(term=lb0->getnext_outputList();term!=0;term=lb0->getnext_outputList())
      term->get_net()->sort_sinkVector();
    lb0->remove_outputListItr();

    lb0->create_inputListItr();
    for(term=lb0->getnext_inputList();term!=0;term=lb0->getnext_inputList())
      term->get_net()->sort_sinkVector();
    lb0->remove_inputListItr();
  }

  if(lb1){
    lb1->create_outputListItr();
    for(term=lb1->getnext_outputList();term!=0;term=lb1->getnext_outputList())
      term->get_net()->sort_sinkVector();
    lb1->remove_outputListItr();

    lb1->create_inputListItr();
    for(term=lb1->getnext_inputList();term!=0;term=lb1->getnext_inputList())
      term->get_net()->sort_sinkVector();
    lb1->remove_inputListItr();
  }

}


void makeMove(hsraBlock** block0,hsraBlock** block1,int* block_type,int iter){
  Terminal* term;
  logicBlock* lb0;logicBlock* lb1;

  lb0=(*block0)->get_currBlock();
  lb1=(*block1)->get_currBlock();

#if defined(PLACER_DEBUG)
  cout<<endl<<endl<<"IN MAKE MOVE...";
#endif

  /* first, rip up the entire net that is sourced at a logic block. also, rip up the */
  /* branch of each net that has a sink at the logic block. the occupancy of the nodes */  
  /* in the ripped-up net segments are accordingly decremented. */
  ripRoutes(lb0,lb1);

  /* clear the terminal lists of the SOURCE and SINK routing nodes in block0 and block1 */
  (*block0)->get_sourceNode()->clear_terminalList();
  (*block0)->get_sinkNode()->clear_terminalList();
  (*block1)->get_sourceNode()->clear_terminalList();
  (*block1)->get_sinkNode()->clear_terminalList();

  /* translate/swap lb0 and/or lb1 */
  placeBlock(lb0,*block1,*block_type);
  if(lb1)
    placeBlock(lb1,*block0,*block_type);
  else
    (*block0)->set_currBlock(0);

  /* now, re-route the nets that connect to lb0 and lb1 */
  if(lb0){
    /* route the output nets of lb0 */
    lb0->create_outputListItr();
    for(term=lb0->getnext_outputList();term!=0;term=lb0->getnext_outputList()){
      /* only consider those nets that have not already been ripped up */
      routeNet(iter,term->get_net(),1,false);
      /* this net has now been counted */
      term->get_net()->set_isCounted(true);
    }
    lb0->remove_outputListItr();
  }

  /* now, re-route the nets that connect to lb0 and lb1 */
  if(lb1){
    /* route the output nets of lb0 */
    lb1->create_outputListItr();
    for(term=lb1->getnext_outputList();term!=0;term=lb1->getnext_outputList()){
      /* only consider those nets that have not already been ripped up */
      routeNet(iter,term->get_net(),1,false);
      /* this net has now been counted */
      term->get_net()->set_isCounted(true);
    }
    lb1->remove_outputListItr();
  }

  if(lb0){
    /* now, route the input terminals of lb0 */ 
    lb0->create_inputListItr();
    for(term=lb0->getnext_inputList();term!=0;term=lb0->getnext_inputList()){
      if(term->get_net()->get_isCounted()==false)
        routeSink(iter,term->get_net(),term->get_rNode(),1);
    }
    lb0->remove_inputListItr();
  }

  if(lb1){
    /* now, route the input terminals of lb0 */ 
    lb1->create_inputListItr();
    for(term=lb1->getnext_inputList();term!=0;term=lb1->getnext_inputList()){
      if(term->get_net()->get_isCounted()==false)
        routeSink(iter,term->get_net(),term->get_rNode(),1);
    }
    lb1->remove_inputListItr();
  }

  /* now that the nets have been ripped up, reset the output nets of lb0 and lb1 */
  if(lb0){
    lb0->create_outputListItr();
    for(term=lb0->getnext_outputList();term!=0;term=lb0->getnext_outputList())
      term->get_net()->set_isCounted(false);
    lb0->remove_outputListItr();
  }
  if(lb1){
    lb1->create_outputListItr();
    for(term=lb1->getnext_outputList();term!=0;term=lb1->getnext_outputList())
      term->get_net()->set_isCounted(false);
    lb1->remove_outputListItr();
  }

#if defined(PLACER_DEBUG)
  cout<<endl<<"EXITING MAKE MOVE...";
#endif

}


/***
* This function rips up the entire source net, and the branch of each net that sinks at
* lb0 and/or lb1. Occupancies are appropriately decremented.
***/  
void ripRoutes(logicBlock* lb0,logicBlock* lb1){
  Terminal* term;

#if defined(PLACER_DEBUG)
  cout<<endl<<endl<<"IN RIP ROUTES...";
#endif

  if(lb0){
    /* loop thru the output nets of lb0 and delete route trees */
    lb0->create_outputListItr();
    for(term=lb0->getnext_outputList();term!=0;term=lb0->getnext_outputList()){
      /* only consider those nets that have not already been ripped up */
      term->get_net()->rip_rTree();
      /* this net has now been counted */
      term->get_net()->set_isCounted(true);
    }
    lb0->remove_outputListItr();
  }

  if(lb1){
    /* loop thru the output nets of lb1 and delete route trees */
    lb1->create_outputListItr();
    for(term=lb1->getnext_outputList();term!=0;term=lb1->getnext_outputList()){
      /* only consider those nets that have not already been ripped up */
      term->get_net()->rip_rTree();
      /* this net has now been counted */
      term->get_net()->set_isCounted(true);
    }
    lb1->remove_outputListItr();
  }

  if(lb0){
    /* now, loop thru the input nets of lb0 and rip up the branch that connects */
    /* lb0 to the route tree of each input net */
    lb0->create_inputListItr();
    for(term=lb0->getnext_inputList();term!=0;term=lb0->getnext_inputList()){
      if(term->get_net()->get_isCounted()==false)
        term->get_net()->rip_rTreeBranch(term->get_rNode());
    }
    lb0->remove_inputListItr();
  }

  if(lb1){
    /* now, loop thru the input nets of lb1 and rip up the branch that connects */
    /* lb1 to the route tree of each input net */
    lb1->create_inputListItr();
    for(term=lb1->getnext_inputList();term!=0;term=lb1->getnext_inputList()){
      if(term->get_net()->get_isCounted()==false)
        term->get_net()->rip_rTreeBranch(term->get_rNode());
    }
    lb1->remove_inputListItr();
  }

  /* now that the nets have been ripped up, reset the output nets of lb0 and lb1 */
  if(lb0){
    lb0->create_outputListItr();
    for(term=lb0->getnext_outputList();term!=0;term=lb0->getnext_outputList())
      term->get_net()->set_isCounted(false);
    lb0->remove_outputListItr();
  }
  if(lb1){
    lb1->create_outputListItr();
    for(term=lb1->getnext_outputList();term!=0;term=lb1->getnext_outputList())
      term->get_net()->set_isCounted(false);
    lb1->remove_outputListItr();
  }

#if defined(PLACER_DEBUG)
  cout<<endl<<endl<<"EXITING RIP ROUTES...";
#endif

}


/***
* This function checks to see if lb is a movable logic block. The
* reason it is here is because there are 'global' nets that are 
* sourced at input blocks that have no sinks (e.g. clock, reset etc)
***/
bool movable_block(logicBlock* lb){
  Net* net;

  if(lb->get_isInput() || lb->get_isOutput())
    return false;
  else{
    if(lb->pos>=FPGA->lsize)
      return false;
    else
      return true;
  }

}


/***
* This function generates an anneal move. When the function is done executing
* block0 and block1 point to the 2 FPGA locations that will be involved in the
* move. block0 always points to a location that holds a logic or IO block, while 
* block1 may or may not be currently occupied. block_type indicates whether the 
* involved blocks are CLBs or IOBs. 
***/
void generateMove(hsraBlock** block0,hsraBlock** block1,int* block_type){
  int num_clbs=FPGA->get_numCLBs();
  int num_iobs=FPGA->get_numIOBs();
  int num_els=netlist.get_numlogicBlocks()+netlist.get_numInputs()+netlist.get_numOutputs();
  int pos0,pos1;
  logicBlock* lb1;



  /***** WINDOWING HERE *****/



  logicBlock* lb0=netlist.get_logicBlock(lrand48()%num_els);
  while(movable_block(lb0)==false){
    /* randomly select a logic or IO block */
    lb0=netlist.get_logicBlock(lrand48()%num_els);
  }

  /* record its current position in the HSRA */
  pos0=lb0->pos;

#if defined(PLACER_DEBUG)
  cout<<endl<<endl<<"LOGIC BLOCK "<<lb0->get_name()<<" AT COL="<<col0<<" AND ROW="<<row0;
#endif

  if(lb0->get_isLogic())
    *block0=FPGA->get_clbAt(pos0);
  else
    *block0=FPGA->get_iobAt(pos0);

  if(lb0->get_isLogic()!=true){
    *block_type=0;

    /* lb0 is an IO block...randomly select another IO location */
    while(1){
      pos1=lrand48()%num_iobs;
      /* check to see whether randomly generated IO location is the same as the other */ 
      if(pos1==pos0)
        continue;
      else{
	*block1=FPGA->get_iobAt(pos1);
	lb1=(*block1)->get_currBlock();
	if(lb1){
	  if(movable_block(lb1)==false)
	    continue;
	}
        break;
      }
    }

#if defined(PLACER_DEBUG)
    if(lb1 = (*block1)->get_currBlock())
      cout<<endl<<endl<<"IO BLOCK "<<lb1->get_name()<<" AT POSITION "<<lb1->pos;
#endif

  }

  else{
    *block_type=1;

    /* lb0 is a logic block...randomly select another CLB location */
    while(1){
      pos1=lrand48()%num_clbs;
      /* check to see whether randomly generated CLB location is the same as the other */ 
      if(pos1==pos0 || pos1>=FPGA->lsize)
        continue;
      else{
	*block1=FPGA->get_clbAt(pos1);
	lb1=(*block1)->get_currBlock();
	if(lb1){
	  if(movable_block(lb1)==false)
	    continue;
	}
        break;
      }
    }

#if defined(PLACER_DEBUG)
    if(lb1 = (*block1)->get_currBlock())
      cout<<endl<<endl<<"LOGIC BLOCK "<<lb1->get_name()<<" AT POSITION "<<lb1->pos;
#endif

  }

}


/* return the smaller of two floating point numbers */
double flomin(double x,double y){
  double diff=y-x;
  if(diff<-1.0e-15)
    return y;
  else
    return x;
}

void printNetsToFile(FILE* fp){
  int wire_cost=0;int sharing_cost=0;

  netlist.create_netListItr();
  for(Net* net=netlist.getnext_netList();net!=0;net=netlist.getnext_netList()){

    if(net->sinkVector==0)
      continue;

    net->print_Info(fp);
    net->print_rTree(fp);
    wire_cost+=net->get_rNodeCount();
    fprintf(fp,"\nWIRE COST = %d",wire_cost);
  }
  netlist.remove_netListItr();

  sharing_cost=FPGA->countSharing(fp);

  fprintf(fp,"\n\n^^^^^^^^^^^ FINAL WIRE COST = %d ^^^^^^^^^^^^",wire_cost);
  fprintf(fp,"\n^^^^^^^^^^^ FINAL SHARING COST = %d ^^^^^^^^^^^^\n\n",sharing_cost);

}

void printNetsAndRecalcCosts(){
  int wire_cost=0;int sharing_cost=0;

  netlist.create_netListItr();
  for(Net* net=netlist.getnext_netList();net!=0;net=netlist.getnext_netList()){

    if(net->sinkVector==0)
      continue;

    net->print_Info();
    net->print_rTree();
    wire_cost+=net->get_rNodeCount();
    cout<<endl<<"WIRE COST = "<<wire_cost;
  }
  netlist.remove_netListItr();

  sharing_cost=FPGA->countSharing(2);

  cout<<endl<<endl<<"^^^^^^^^^^^ FINAL WIRE COST = "<<wire_cost<<" ^^^^^^^^^^^^";
  cout<<endl<<"^^^^^^^^^^^ FINAL SHARING COST = "<<sharing_cost<<" ^^^^^^^^^^^^";
  cout<<endl<<endl;

}


/***
* This function routes each net in the netlist. The parameter 'iter'
* will be used to determine the cost (Pathfinder) of routing nodes. 
***/
void routeNets(int iter,bool pathfinder_pshare){
  int net_ctr=0;

  netlist.create_netListItr();
  for(Net* net=netlist.getnext_netList();net!=0;net=netlist.getnext_netList()){

    if(net->sinkVector==0)
      continue;

    if(iter==1){
      routeNet(iter,net,0,pathfinder_pshare);
      /*        net->print_Info(); */
      /*        net->print_rTree(); */
      wireCost+=net->get_rNodeCount();
      /*       cout<<endl<<"WIRE COST = "<<wireCost; */
    }
    else{
      net->rip_rTree();
      routeNet(iter,net,1,pathfinder_pshare);
    }

#if defined(ROUTER_DEBUG)
    net->print_Info();
    net->print_rTree();
#endif

  }
  netlist.remove_netListItr();

  sharingCost=FPGA->countSharing(0);

}

int calcWireCost(){
  int wire_cost=0;

  netlist.create_netListItr();
  for(Net* net=netlist.getnext_netList();net!=0;net=netlist.getnext_netList()){
    if(net->sinkVector==0)
      continue;
    wire_cost+=net->get_rNodeCount();
  }
  netlist.remove_netListItr();

  return wire_cost;

}

/***
* This function finds a route for the net pointed to by 'net'. It assumes that all
* routingNode and Net objects have been cleaned up and initialized for routing.
* This cleanup MUST be separately handled by another routine. Specifically, the 
* previous routing tree should have been ripped up, and routing node members (isExplored,
* occupancy,presentSharing,pathCost,predecessor and inrTree) should have been 
* accordingly changed. 
***/
void routeNet(int iter,Net* net,int change_costs,bool pathfinder_pshare){
  routingNode* rnode;routingNode* nb_rnode;
  routingEdge* nb_edge;
  double path_cost,nb_path_cost,new_path_cost,heur_estimate;
  Terminal* terminal;
  priority_queue<pqNode>* PQ;
  treeNode* tree_node;
  list<routingNode*> cleanupList;
  list<routingNode*>::iterator clean_itr;

  /* new root tree node. initialize it to source routing node */
  net->init_rTreeRoot();
  PQ=new priority_queue<pqNode>;

  /* get source routing node and add to priority queue at path cost 0. */
  /* include the source in the routing tree, and increment occupancy. */
  terminal=net->get_source();
  rnode=terminal->get_rNode();
  PQ->push(pqNode(rnode,0.,0,0));
  rnode->set_inrTree(true);
  rnode->set_occupancy(rnode->get_occupancy()+1);
  if(change_costs==1){
    if(rnode->get_occupancy()>rnode->get_capacity())
      ERROR("number of signals occupying source node exceeds capacity...");
  }

#if defined(ROUTER_DEBUG)
  cout<<endl<<endl<<"<<<<<<<<<============ BEGIN ROUTER DEBUG FOR "<<net->get_name();
  cout<<"============>>>>>>>>>";
  cout<<endl<<"SOURCE LOGIC BLOCK OF THIS NET IS "<<terminal->get_lBlock()->get_name();
  cout<<" AND ITS OUTPUT IS MAPPED TO "<<rnode->get_name();
#endif

  /* start exploration */
  routingNode* sink_rnode;
  int sink_count=0;
  for(;;){
    sink_rnode=net->sinkVector->at(sink_count)->rNode;
    
    /* loop until an unexplored routing node is obtained */
    while(1){
      if(PQ->empty()){
	cout<<endl<<endl;
	net->print_Info();
        ERROR("priority queue is empty before net routing completed...");
      }
      rnode=(PQ->top()).rNode;
      if(rnode->isExplored==false){
        /* found unexplored routing node...mark it explored,set its pathcost and */
        /* predecessor routing node and break */
        rnode->isExplored=true;
	rnode->pathCost=(PQ->top()).pathCost;
	rnode->predecessor=(PQ->top()).predNode;
        cleanupList.push_back(rnode);  //this list holds routing nodes that need a clean-up
        PQ->pop();
        break;
      }
      PQ->pop();
    }

#if defined(ROUTER_DEBUG)
    cout<<endl<<"POPPED "<<rnode->get_name()<<" AT PATH COST "<<rnode->get_pathCost();
    cout<<" AND PREDECESSOR NODE ";
    if(rnode->get_predecessor())
      cout<<rnode->get_predecessor()->get_name();
#endif

    if(rnode->type==SINK){
      /* found a sink terminal of net */
      if(rnode==sink_rnode && rnode->isRouted==false){
#if defined(ROUTER_DEBUG)
        cout<<endl<<endl<<"++++++++++++ FOUND A SINK; WILL NOW ENTER BACKTRACE ++++++++++++";
#endif
        tree_node=backTrace(net,rnode,iter,change_costs);

	if(sink_count<(net->sinkVector->size()-1)){
	  /* will need to delete the old PQ first, and create a new one */
	  delete PQ;PQ=new priority_queue<pqNode>;

/* 	  net->print_rTree(); */

	  /* now reinit the PQ to the current routing tree ala VPR's directed search */ 
	  updatePQ(PQ,(net->rTree).root,net->sinkVector->at(sink_count+1)->rNode);
	}

        rnode->isRouted=true;

        /* clean up all routing nodes */
        for(clean_itr=cleanupList.begin();clean_itr!=cleanupList.end();clean_itr++){
	  if((*clean_itr)->inrTree==false){
	    (*clean_itr)->isExplored=false;
	    (*clean_itr)->predecessor=0;
	    (*clean_itr)->pathCost=HUGE_NUMBER;
	    (*clean_itr)->isRouted=false;
	  }
	}

	sink_count++;

	if(sink_count==net->sinkVector->size()){
	  delete PQ;
	  break;
	}
	else
	  continue;

      }
    }

    /* loop thru the fanout neighbor nodes of rnode */ 
    rnode->create_fanoutListItr();
    for(nb_edge=rnode->getnext_fanoutList();nb_edge!=0;nb_edge=rnode->getnext_fanoutList()){
      nb_rnode=nb_edge->destNode;

      /* A* algorithm */
      heur_estimate=ALPHA*cost_to_target(nb_rnode,sink_rnode);

#if defined(ROUTER_DEBUG)
      cout<<endl<<"LOOKING AT NEIGHBOR NODE "<<nb_rnode->get_name();
#endif

      if(nb_rnode->isExplored==false){
        path_cost=rnode->pathCost;
	new_path_cost=path_cost+nb_rnode->returnCost(iter,pathfinder_pshare);
	PQ->push(pqNode(nb_rnode,new_path_cost,heur_estimate,rnode));
#if defined(ROUTER_DEBUG)
	cout<<endl<<"\tADDED NODE "<<nb_rnode->get_name()<<" AT PATH COST "<<
	  new_path_cost<<" AND PREDECESSOR "<<rnode->get_name();
#endif

      }

    }
    rnode->remove_fanoutListItr();

  }
  /* end exploration */

  /* clean up all routing nodes that have been affected by the exploration */
  for(clean_itr=cleanupList.begin();clean_itr!=cleanupList.end();clean_itr++){
    (*clean_itr)->isExplored=false;
    (*clean_itr)->inrTree=false;
    (*clean_itr)->predecessor=0;
    (*clean_itr)->pathCost=HUGE_NUMBER;
    (*clean_itr)->isRouted=false;
  }

}

/***
* This function finds a route to sink_rnode from net's route tree. It is assumed here
* that there is an existing route tree that connects all sinks EXCEPT sink_rnode. Routing 
* node members (isExplored,occupancy,presentSharing,pathCost,predecessor and inrTree) 
* should be appropriately initialized.  
***/
void routeSink(int iter,Net* net,routingNode* sink_rnode,int change_costs){
  routingNode* rnode;routingNode* nb_rnode;
  routingEdge* nb_edge;
  double path_cost,nb_path_cost,new_path_cost,heur_estimate;
  Terminal* terminal;
  priority_queue<pqNode> PQ;
  treeNode* tree_node;
  list<routingNode*> cleanupList;

  /* init the PQ to the existing route tree of net */
  initPQ(PQ,(net->get_rTree())->get_root(),sink_rnode);

  /* start exploration */
  while(1){
    
    /* loop until an unexplored routing node is obtained */
    while(1){
      if(PQ.empty())
	ERROR("priority queue is empty before net routing completed...");

      rnode=PQ.top().rNode;
      if(rnode->isExplored==false){
	/* found unexplored routing node...mark it explored,set its pathcost and */
	/* predecessor routing node and break */
	rnode->isExplored=true;
	rnode->pathCost=PQ.top().pathCost;
	rnode->predecessor=PQ.top().predNode;
	cleanupList.push_back(rnode);  //this list holds routing nodes that need a clean-up
	PQ.pop();
	break;
      }

      PQ.pop();

    }

#if defined(ROUTER_DEBUG)
    cout<<endl<<"POPPED "<<rnode->get_name()<<" AT PATH COST "<<rnode->get_pathCost();
    cout<<" AND PREDECESSOR NODE ";
    if(rnode->get_predecessor())
      cout<<rnode->get_predecessor()->get_name();
#endif

    if(rnode->type==SINK){
      /* found the sink terminal that we are looking for */
      if(rnode==sink_rnode){
#if defined(ROUTER_DEBUG)
	cout<<endl<<endl<<"++++++++++++ FOUND SINK; WILL NOW ENTER BACKTRACE ++++++++++++";
#endif

	backTrace(net,rnode,iter,change_costs);

	/* clean up all routing nodes that have been affected by the exploration */
	list<routingNode*>::iterator clean_itr;
	for(clean_itr=cleanupList.begin();clean_itr!=cleanupList.end();clean_itr++){
	  (*clean_itr)->isExplored=false;
	  (*clean_itr)->inrTree=false;
	  (*clean_itr)->predecessor=0;
	  (*clean_itr)->pathCost=HUGE_NUMBER;
	  (*clean_itr)->isRouted=false;
	}

	/* now clean up all the routing nodes that are in the new route tree */
	(net->get_rTree())->clean_rNodes(net->get_rTree()->get_root());

	/* exploration ends...return */
	return;

      }
    }

    /* loop thru the fanout neighbor nodes of rnode */ 
    rnode->create_fanoutListItr();
    for(nb_edge=rnode->getnext_fanoutList();nb_edge!=0;nb_edge=rnode->getnext_fanoutList()){
      nb_rnode=nb_edge->destNode;

      /* A* algorithm */
      heur_estimate=ALPHA*cost_to_target(nb_rnode,sink_rnode);

#if defined(ROUTER_DEBUG)
      cout<<endl<<"LOOKING AT NEIGHBOR NODE "<<nb_rnode->get_name();
#endif

      if(nb_rnode->isExplored==false){
	path_cost=rnode->pathCost;
	nb_path_cost=nb_rnode->pathCost;
	new_path_cost=path_cost+nb_rnode->returnCost(iter,false);
	PQ.push(pqNode(nb_rnode,new_path_cost,heur_estimate,rnode));
#if defined(ROUTER_DEBUG)
	cout<<endl<<"\tADDED NODE "<<nb_rnode->get_name()<<" AT PATH COST "<<
	  new_path_cost<<" AND PREDECESSOR "<<rnode->get_name();
#endif
      }

    }
    rnode->remove_fanoutListItr();

  }

}

void initPQ(priority_queue<pqNode>& PQ,treeNode* tree_node,routingNode* new_sink_rnode){
  treeNode* t=tree_node;
  treeNode* child;
  double heur_estimate;

  t->get_rNode()->set_inrTree(true);

  heur_estimate=ALPHA*cost_to_target(t->get_rNode(),new_sink_rnode);
  if(t->get_rNode()->get_type()!=SINK && t->get_rNode()->get_type()!=INPUT_PIN){
    if(t->get_parent())
      PQ.push(pqNode(t->get_rNode(),0.,heur_estimate,t->get_parent()->get_rNode()));
    else
      PQ.push(pqNode(t->get_rNode(),0.,heur_estimate,0));
    t->get_rNode()->set_pathCost(0.);
  }

  t->create_childListItr();
  for(child=t->getnext_childList();child!=0;child=t->getnext_childList())
    initPQ(PQ,child,new_sink_rnode);
  t->remove_childListItr();

}

void updatePQ(priority_queue<pqNode>* PQ,treeNode* tree_node,routingNode* new_sink_rnode){
  treeNode* t=tree_node;
  treeNode* child;
  double heur_estimate;

  if(t->get_rNode()==0)
    ERROR("something fishy with the backTrace operation here...");
  if(t->get_parent()){
    if(t->get_parent()->get_rNode()!=t->get_rNode()->get_predecessor())
      ERROR("parent mismatch here...");
  }

  if(t->get_rNode()->get_type()!=SINK && t->get_rNode()->get_type()!=INPUT_PIN){
    /* A* algorithm */

    heur_estimate=ALPHA*cost_to_target(t->get_rNode(),new_sink_rnode);
    PQ->push(pqNode(t->get_rNode(),0.,heur_estimate,t->get_rNode()->get_predecessor()));
    t->get_rNode()->set_pathCost(0.);
    t->get_rNode()->set_inrTree(true);
    t->get_rNode()->set_isExplored(false);
  }

  t->create_childListItr();
  for(child=t->getnext_childList();child!=0;child=t->getnext_childList())
    updatePQ(PQ,child,new_sink_rnode);
  t->remove_childListItr();

}

treeNode* backTrace(Net* net,routingNode* sink_rnode,int iter,int change_costs){
  treeNode* curr_tree_node;
  treeNode* prev_tree_node=0;
  routingNode* curr_node;

  /* start backtracing from the newly found sink to a node in the routing tree */
  for(curr_node=sink_rnode;curr_node->get_inrTree()==false;
      curr_node=curr_node->get_predecessor()){

    /* create a new tree node for the current routing node on the backtraced path */
    curr_tree_node=new treeNode(curr_node,curr_node->get_pathCost());

#if defined(ROUTER_DEBUG)
  cout<<endl<<"\tADDED "<<curr_node->get_name()<<" TO ROUTING TREE";
#endif

/*     cout<<endl<<"\tADDED "<<curr_node->get_name()<<" TO ROUTING TREE"; */

    if(prev_tree_node!=0){
      /* hook up parent and child tree nodes */
      curr_tree_node->addto_childList(prev_tree_node);
      prev_tree_node->set_parent(curr_tree_node);
    }

    curr_node->set_occupancy(curr_node->get_occupancy()+1);
    prev_tree_node=curr_tree_node;

    /* change the sharing and wire costs to reflect the addition of this node */
    if(change_costs==1){
      if(curr_node->get_occupancy()>curr_node->get_capacity())
        sharingCost+=1;
      if(curr_node->get_type()==BUS_OR_WIRE)
        wireCost+=1;
    }

  }

  /* curr_node now points to a routing node that is present in the route tree. */
/*   cout<<endl<<endl<<"NEED TO FIND "<<curr_node->get_name()<<" IN ROUTE TREE"; */

  /* find the tree node corresponding to curr_node */
  curr_tree_node=net->find_treeNode(curr_node);

  /* connect the new tree segment to the existing route tree */
  curr_tree_node->addto_childList(prev_tree_node);
  prev_tree_node->set_parent(curr_tree_node);

#if defined(ROUTER_DEBUG)
  cout<<endl<<endl<<"LEAVING BACKTRACE...";
#endif

/*   cout<<endl<<endl<<"FOUND TREE NODE "<<curr_tree_node->get_rNode()->get_name(); */

  /* return the tree node that joins the new tree segment to the existing route tree */
  return prev_tree_node; 

}

/***
* This function places the logic block lb on FPGA block b. The parameter
* t indicates whether this is a HSRA_CLB or HSAR_IOB.
* Note that we assume here that there is no other logic block already 
* mapped to the FPGA block b...the calling function MUST check occupancy.
***/
void placeBlock(logicBlock* lb,hsraBlock* b,int t){
  hsraBlock* clb;hsraBlock* iob;
  routingNode* rnode;
  Terminal* term;

  if(lb==0 || b==0)
    ERROR("NULL parameters passed to function...");
  if(t!=0 && t!=1){
    cout<<endl<<endl<<"BLOCK TYPE = "<<t;
    ERROR("invalid FPGA block type...");
  }

  if(t==1){
    /* We have a CLB */
    clb=b;

/*      if(clb->get_currBlock()) */
/*        ERROR("hey, this CLB is already taken...");  */

    clb->set_currBlock(lb);
    lb->pos=clb->get_position();

    /* first map the input nets of lb to the sink node of clb */
    rnode=clb->get_sinkNode();
    lb->create_inputListItr();
    for(term=lb->getnext_inputList();term!=0;term=lb->getnext_inputList()){
      rnode->addto_terminalList(term);
      term->set_rNode(rnode);
    }
    lb->remove_inputListItr();

    /* now map the output nets of lb to the source node of clb */
    rnode=clb->get_sourceNode();
    lb->create_outputListItr();
    for(term=lb->getnext_outputList();term!=0;term=lb->getnext_outputList()){
      rnode->addto_terminalList(term);
      term->set_rNode(rnode);
    }
    lb->remove_outputListItr();

  }
  else{
    /* We have an IOB */
    iob=b;

/*      if(iob->getfrom_currBlockArray(io_index)) */
/*        ERROR("hey, this IOB is already taken..."); */

    iob->set_currBlock(lb);
    lb->pos=iob->get_position();

    if(lb->get_isOutput()){
      /* We have an output pad */
      rnode=iob->get_sinkNode();
      lb->create_inputListItr();
      term=lb->getnext_inputList();
      rnode->addto_terminalList(term);
      term->set_rNode(rnode);
      lb->remove_inputListItr();
    }
    else{
      /* We have an input pad */
      rnode=iob->get_sourceNode();
      lb->create_outputListItr();
      term=lb->getnext_outputList();
      rnode->addto_terminalList(term);
      term->set_rNode(rnode);
      lb->remove_outputListItr();
    }

  }

}


void initExternalPlacement(){
  logicBlock* lb;hsraBlock* clb;
  Terminal* term;

  for(int i=0;i<placeVec->size();i++){
    if(placeVec->at(i)!=0){
      lb=netlist.found_logicBlock(placeVec->at(i));
      clb=FPGA->clbArray->at(i);
      if(i>=FPGA->lsize)
	ERROR("exceeded lsize during placement...");
      placeBlock(lb,clb,1);
    }
  }

  netlist.create_lbListItr();
  for(lb=netlist.getnext_lbList();lb!=0;lb=netlist.getnext_lbList()){
    if(lb->get_isInput()){
      lb->create_outputListItr();
      term=lb->getnext_outputList();
      FPGA->iobSource->addto_terminalList(term);
      term->set_rNode(FPGA->iobSource);
      lb->remove_outputListItr();
    }
    if(lb->get_isOutput()){
      lb->create_inputListItr();
      term=lb->getnext_inputList();
      FPGA->iobSink->addto_terminalList(term);
      term->set_rNode(FPGA->iobSink);
      lb->remove_inputListItr();
    }
  }
  netlist.remove_lbListItr();

}


/***
* This function will create a random initial placement of the netlist on the
* architecture pointed to by FPGA
***/
void initPlacement(){
  int num_clbs=FPGA->get_numCLBs();
  int num_iobs=FPGA->get_numIOBs();
  int pos;
  int i;
  hsraBlock* iob;hsraBlock* clb;

  netlist.create_lbListItr();
  for(logicBlock* lb=netlist.getnext_lbList();lb!=0;lb=netlist.getnext_lbList()){

    if(lb->get_isLogic()){
      while(1){
	pos=lrand48()%FPGA->lsize;
	clb=FPGA->get_clbAt(pos);
	if(clb->get_currBlock())
	  continue;
	else{
	  placeBlock(lb,clb,1);
	  break;
	}
      }
    }
    else{
      while(1){
	pos=lrand48()%num_iobs;
	iob=FPGA->get_iobAt(pos);
	if(iob->get_currBlock())
	  continue;
	else{
	  placeBlock(lb,iob,0);
	  break;
	}
      }
    }

  }
  netlist.remove_lbListItr();

}

void printHSRAPlacement(FILE* fp){

  for (int i=0;i<FPGA->get_numCLBs();i++){
    if(FPGA->clbArray->at(i)->get_currBlock())
      fprintf(fp,"%d %s\n",i,FPGA->clbArray->at(i)->get_currBlock()->get_name());
  }

}

/***
* This function prints the current placement of the netlist on the FPGA
***/
void printPlacement(){
  hsraBlock* block;
  routingNode* rnode;
  Terminal* term;

  cout<<endl<<endl;
  cout<<endl<<"<<<<<<<<<<<<<<<<<<<<< BEGIN PRINT PLACEMENT >>>>>>>>>>>>>>>>>>>>>";

  netlist.create_lbListItr();
  for(logicBlock* lb=netlist.getnext_lbList();lb!=0;lb=netlist.getnext_lbList()){
    if(lb->get_isLogic()){
      cout<<endl<<lb->get_name()<<" AT POSITION ["<<lb->pos<<"]";
      block=FPGA->get_clbAt(lb->pos);

      /* CORRESPONDENCE CHECK */
      if(lb!=block->get_currBlock())
	ERROR("correspondence violation in placement phase...");

      /* print the nets that connect to the inputs of the logic block mapped to tile */
      rnode=block->get_sinkNode();
      cout<<endl<<"SINK NODE IS "<<rnode->get_name()<<". NETS MAPPED TO SINK NODE -->";
      rnode->create_terminalListItr();
      for(term=rnode->getnext_terminalList();term!=0;term=rnode->getnext_terminalList()){
	/* CORRESPONDENCE CHECK */
	if(term->get_rNode()!=rnode)
	  ERROR("correspondence violation in placement phase...");
	cout<<endl<<"\t"<<term->get_net()->get_name();
      }
      rnode->remove_terminalListItr();

      /* print the nets that connect to the outputs of the logic block mapped to tile */
      rnode=block->get_sourceNode();
      cout<<endl<<"SOURCE NODE IS "<<rnode->get_name()<<". NETS MAPPED TO SOURCE NODE -->"; 
      rnode->create_terminalListItr();
      for(term=rnode->getnext_terminalList();term!=0;term=rnode->getnext_terminalList()){
	/* CORRESPONDENCE CHECK */
	if(term->get_rNode()!=rnode)
	  ERROR("correspondence violation in placement phase...");
	cout<<endl<<"\t"<<term->get_net()->get_name();
      }
      rnode->remove_terminalListItr();

      cout<<endl;
    }
  }
  netlist.remove_lbListItr();

  cout<<endl<<"<<<<<<<<<<<<<<<<<<<<< END PRINT PLACEMENT >>>>>>>>>>>>>>>>>>>>>";

}

/***
 * This routine iterates thru each input of the blif 
 * netlist. For each input, new logic block and net 
 * data_structs are created and stored.
 ***/
void processBlifInputs(){
  blifClb* input_block;
  blifNet* net;

  list<char*>::iterator s_itr = stringList.begin();
  for(;s_itr!=stringList.end();s_itr++){

#if defined(PARSER_DEBUG)
    cout<<endl<<"PROCESSING INPUT WITH NAME ";
    cout<<*s_itr<<endl;
#endif

    /* create a new blif input block */
    input_block = new blifClb;
    /* give it a name */
    input_block->set_name(*s_itr);
    /* make it an input block */
    input_block->set_isInputBlock(true);
    /* add to blif_netlist */
    blif_netlist.addto_blifClbList(input_block);
    /* look for the net, update data_structs */
    net = find_blifNet(*s_itr);
    input_block->addto_outputList(net);
    net->set_source(input_block);

  }

}

/***
 * This routine iterates thru each output of the blif 
 * netlist. For each output, new logic block and net 
 * data_structs are created and stored.
 ***/
void processBlifOutputs(){
  blifClb* output_block;
  blifNet* net;
  char oname [MAX_STR_LEN];

  list<char*>::iterator s_itr = stringList.begin();
  for(;s_itr!=stringList.end();s_itr++){

#if defined(PARSER_DEBUG)
    cout<<endl<<"PROCESSING OUTPUT WITH NAME ";
    cout<<*s_itr<<endl;
#endif

    /* create a new blif output block */
    output_block = new blifClb;
    /* give it a name */
    sprintf(oname,"%s_outBlock",*s_itr);
    output_block->set_name(oname);
    /* make it an output block */
    output_block->set_isOutputBlock(true);
    /* add to blif_netlist */
    blif_netlist.addto_blifClbList(output_block);
    /* look for the net, update data_structs */
    net = find_blifNet(*s_itr);
    output_block->addto_inputList(net);
    net->addto_sinkList(output_block);

  }

}

void processBlifClb(){
  blifClb* logic_block;
  blifNet* net;

  /* create a new blif logic block */
  logic_block = new blifClb;
  logic_block->set_isClb(true);

  /* add to blif_netlist */
  blif_netlist.addto_blifClbList(logic_block);

  list<char*>::iterator s_itr = stringList.begin();
  for(int i=1;s_itr!=stringList.end();s_itr++,i++){
    if(i==stringList.size()){
      logic_block->set_name(*s_itr);
      net = find_blifNet(*s_itr);
      logic_block->addto_outputList(net);
      net->set_source(logic_block);
    }
    else{
      net = find_blifNet(*s_itr);
      logic_block->addto_inputList(net);
      net->addto_sinkList(logic_block);
    }

  }

#if defined(PARSER_DEBUG)
    cout<<endl<<"PROCESSING CLB WITH NAME ";
    cout<<logic_block->get_name()<<endl;
#endif

}

/***
 * This routine creates and stores a latch logic block.
 * input_str is the name of the input net of the latch,
 * and output_str is the name of the output net.
 ***/
void processBlifLatch(char* input_str,char* output_str){
  blifClb* latch_block;
  blifNet* net;

  /* create a new blif logic block */
  latch_block = new blifClb;
  latch_block->isLatch=true;
  latch_block->set_name(output_str);

  /* add to blif_netlist */
  blif_netlist.addto_blifClbList(latch_block);

  net=find_blifNet(input_str);
  latch_block->addto_inputList(net);
  net->addto_sinkList(latch_block);

  net=find_blifNet(output_str);
  latch_block->addto_outputList(net);
  net->set_source(latch_block);

}

/***
* Look for a net with name net_name in blif_netlist. If found, return it
* else create a new net, give it a name, add it to blif_netlist and then
* return it.
***/
blifNet* find_blifNet(char* net_name){
  blifNet* return_net;
  if(!net_name)
    ERROR("name of a net cannot be NULL...");
  if(return_net = blif_netlist.found_blifNet(net_name))
    return return_net;
  else{
    return_net = new blifNet;
    return_net->set_name(net_name);
    blif_netlist.addto_blifNetList(return_net);
    return return_net;
  }
}

/***
* Print the entire blif netlist data structure
***/
void print_blifNetlist(){

  cout<<endl<<endl;
  cout<<endl<<"NAME OF BLIF NETLIST IS "<<blif_netlist.get_name();
  cout<<endl<<"NUMBER OF CLBS = "<<blif_netlist.get_numClbs();
  cout<<endl<<"NUMBER OF INPUT BLOCKS = "<<blif_netlist.get_numInputs();
  cout<<endl<<"NUMBER OF OUTPUT BLOCKS = "<<blif_netlist.get_numOutputs();
  cout<<endl<<"NUMBER OF NETS = "<<blif_netlist.get_numNets();
  blif_netlist.print_blifClbList();
  blif_netlist.print_blifNetList();

}

/***
* This function converts the blif netlist to a generic netlist. This is the
* step that converts benchmark netlists of a particular format to a represe
* ntation that the Independence algorithm routines will recognize and
* operate upon.
***/
void convert2GenericNetlist(blifNetlist& n){
  n.convert2Generic(netlist);
}
  
/***
* Print the entire generic netlist data structure
***/
void print_GenericNetlist(){

  cout<<endl<<endl;
  cout<<endl<<"NAME OF NETLIST IS "<<netlist.get_name();
  cout<<endl<<"NUMBER OF LOGIC BLOCKS = "<<netlist.get_numlogicBlocks();
  cout<<endl<<"NUMBER OF INPUT BLOCKS = "<<netlist.get_numInputs();
  cout<<endl<<"NUMBER OF OUTPUT BLOCKS = "<<netlist.get_numOutputs();
  cout<<endl<<"NUMBER OF NETS = "<<netlist.get_numNets();
  netlist.print_lbList();
  netlist.print_netList();

}

/***
* Create an HSRA FPGA object.
***/
void create_hsraFPGA(int num_clbs,int base_width,int num_levels,float growth_rate,int lsize){

  FPGA = new hsraFPGA(num_clbs,base_width,num_levels,((double)(growth_rate)),lsize);
/*   cout<<endl<<endl; */
/*   FPGA->print_info(); */

}

