////////////////////////////////////////////////////////////////////
// logicBlock.h: Class definition of a generic logic block in a
//               netlist.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

/****************************************************************
Class definition of a generic logic block in a netlist. The 
Independence algo will operate on objects of type 'logicBlock'
'Net' and 'Terminal', rather than on class types that are SPECIFIC 
to the benchmark set. It is the user's responsibility to convert 
a netlist in a particular benchmark set to objects of the generic 
classes 'logicBlock', 'Net' and 'Terminal'.
****************************************************************/    
#ifndef LOGIC_BLOCK_H
#define LOGIC_BLOCK_H

#include "error.h"
#include "Terminal.h"

class Terminal;

class logicBlock{
  char name [MAX_STR_LEN];  //name of logic block
  int functionType;  //in case logic blocks implement different functionalities.
                     //for example, a logic block may be an ALU, Multiplier etc in case
                     //the underlying architecture supports heterogenous logic blocks.
                     //different logic block types can be #defined in error.h
  bool isLogic;  //whether this block actually implements logic
  bool isInput;  //whether this block in an input-only IO block
  bool isOutput;  //whether this block is an output-only IO block
  bool isBidirectional;  //whether this block is a bi-directional IO block
  bool isRegister;  //whether this block is latched / registered or not
  list<Terminal*> inputList;  //list of input terminals
  list<Terminal*> outputList;  //list of output terminals

  list<Terminal*>::iterator* interm_itr;  //pointer to inputList iterator
  list<Terminal*>::iterator* outterm_itr;  //pointer to outputList iterator

  int numInputs;  //num input terminals
  int numOutputs;  //num output terminals
  int magicNum;
 public:

  /********** temporary geometric info **********/
  int pos;  // position that this logic block is mapped to

  logicBlock(){
    functionType=NO_TYPE;
    isLogic=isInput=isOutput=isBidirectional=isRegister=false;
    numInputs=numOutputs=0;
    interm_itr=0;outterm_itr=0;
    magicNum=15111;
  }

  ~logicBlock(){
    if(magicNum!=15111)
      ERROR("trying to double delete a node...");
    else
      magicNum=0;
  }

  void create_inputListItr();  //create inputList iterator
  void create_outputListItr();  //create outputList iterator
  void remove_inputListItr();  //remove inputList iterator
  void remove_outputListItr();  //remove outputList iterator
  Terminal* getnext_inputList();  //get next element from inputList
  Terminal* getnext_outputList();  //get next element from outputList

  void set_functionType(int f);  //set the funtionality type of this block
  void set_name(char* n);  //give this block a name
  void set_isLogic(bool u);  //set this if block actually implements logic
  void set_isInput(bool u);  //set this if block is an input-only IO block
  void set_isOutput(bool u);  //set this if block is an output-only IO block
  void set_isBidirectional(bool u);  //set this if block is a bidirectional IO block
  void set_isRegister(bool u);  //set this if this is a registered / latched block
  void addto_inputList(Terminal* t);  //add a terminal to the input terminal list
  void addto_outputList(Terminal* t);  //add a terminal to the output terminal list

  char* get_name();  //get name of this block
  int get_functionType();  // get functionality type of this block
  bool get_isLogic();  //check if this block actually implements logic
  bool get_isInput();  //check if this block is an input-only IO block
  bool get_isOutput();  //check if this block is an output-only IO block
  bool get_isBidirectional();  //check if this block is a bidirectional IO block
  bool get_isRegister();  //check if this block is a registered / latched block
  int get_numInputs();  //get num input terminals
  int get_numOutputs();  //get num output terminals
  void print_Info();  //print details about this logic block

};

#endif


