////////////////////////////////////////////////////////////////////
// routingNode.h: Class that describes a node in the routing graph.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

/****************************************************************
Class definition of a routing node. There is a separate routing
node for each logic block and IO pad terminal. Further, every 
wire or bus segment in the FPGA architecture is also represented 
as a routing node.

We assume the Elmore delay model here. Thus, each wire segment
routing node has to have a resistance and a capacitance.  
****************************************************************/    
#ifndef ROUTING_NODE_H
#define ROUTING_NODE_H

#include "error.h"
#include "Terminal.h"
#include "routingEdge.h"
#include "Net.h"

class routingEdge;
class Terminal;
class Net;

class routingNode{
 public:
  char name [MAX_STR_LEN];  //name of routing node
  int type;  //type of routing node i.e. source, sink, wire/bus, logic block
             //input, or logic block output
  double R,Cin,Cout,C;  //wire routing nodes have a resistance and capacitances
  list<Terminal*> terminalList;  //each source/sink routing node corresponds to one or more
                                 //logic block or IO pad terminals
  list<Terminal*>::iterator* terminal_itr;  //pointer to terminal list iterator
  double baseCost,historyCost,presentSharing;  //Pathfinder cost function parameters
  double pfac;  //tuning parameter for presentSharing...VPR router
  int occupancy;  //num nets currently using this node
  int capacity;  //capacity of this node
  bool isExplored;  //whether this routing node has been explored by a wavefront
  bool inrTree;  //whether this node is in the current route tree of the net being routed 
  double pathCost;  //current cost of path to this node from source of net -- Pathfinder
  list<routingEdge*> faninList;  //list of fanin routing edges
  list<routingEdge*> fanoutList;  //list of fanout routing edges
  list<routingEdge*>::iterator* fanin_itr;  //pointer to fanin list iterator
  list<routingEdge*>::iterator* fanout_itr;  //pointer to fanout list iterator
  routingNode* predecessor;  //the fanin routing node that was used to explore this node
  bool isRouted;  //only valid for sink routing nodes. true if the sink has already been routed.
  bool counted;  //whether this node's sharing has already been included in cost calculations
  bool canUse;  //whether this routing node can be included in the final routing graph
  int magicNum;

  /************ HSRA A* PURPOSES ************/
  list<char> LRList;
  int lrSize;

  int sinkID;
  int clusterNumber;
  vector<double> allSinksTable;

  routingNode(){
    type=NO_TYPE;
    R=Cin=Cout=C=0.;
    presentSharing=0.;
    baseCost=historyCost=1.0;
    occupancy=capacity=0;
    isExplored=false;
    pathCost=HUGE_NUMBER;
    fanin_itr=fanout_itr=0;terminal_itr=0;
    pfac=1.0;
    predecessor=0;
    inrTree=false;
    isRouted=false;
    counted=false;
    magicNum=19111;
    canUse=true;
    sinkID=-1;clusterNumber=-1;
    lrSize=0;
  }
  routingNode(char* n,int t,double r,double cin,double cout,int cap,double bcost){
    if(n==0)
      ERROR("name of routing node cannot be NULL...");
    if(t<SOURCE || t>OUTPUT_PIN)
      ERROR("mysterious routing node type...");
    if(r<0. || cin<0. || cout<0.)
      ERROR("resistance / capacitance cannot be -ve...");
    if(cap<0)
      ERROR("routing node cannot have -ve capacity...");
    if(bcost<0.)
      ERROR("base cost of routing node cannot be -ve...");
    strcpy(name,n);
    type=t;
    R=r;Cin=cin;Cout=cout;C=0.;
    capacity=cap;
    baseCost=bcost;

    presentSharing=0.;
    historyCost=1.0;
    occupancy=0;
    isExplored=false;
    pathCost=HUGE_NUMBER;
    fanin_itr=fanout_itr=0;terminal_itr=0;
    pfac=1.0;
    predecessor=0;
    inrTree=false;
    isRouted=false;
    counted=false;
    magicNum=19111;
    canUse=true;
    sinkID=-1;clusterNumber=-1;
    lrSize=0;
  }
  ~routingNode(){
    if(magicNum!=19111)
      ERROR("trying to double delete a node...");
    else
      magicNum=0;
  }

  void set_name(char* n);  //give this routing node a name
  void set_type(int t);  //set type of routing node
  void set_R(double r);  //set resistance of wire routing node
  void set_Cin(double c);  //set input cap of wire routing node
  void set_Cout(double c);  //set output cap of wire routing node
  void set_predecessor(routingNode* pr);  //set this node's wavefront predecessor

  void addto_terminalList(Terminal* t);  //add a terminal that is mapped to this node
  void create_terminalListItr();  //create a new iterator for terminalList
  void remove_terminalListItr();  //remove an already existing terminalList iterator
  Terminal* getnext_terminalList();  //get next element from terminalList

  void set_baseCost(double b);  //set base cost of this routing node
  void update_historyCost();  //update history cost of this routing node
  void update_pfac();  //update the tuning parameter for presentSharing...VPR router
  void set_occupancy(int occ);  //set the occupancy of this routing node
  void set_capacity(int c);  //set the capacity of this routing node
  void set_isExplored(bool e);  //whether this node has been explored by a wavefront
  void set_inrTree(bool r);//whether this node is in the current route tree of
                             //the net being routed
  bool get_inrTree();//return true if this routing node is in the route tree of a net
  void set_pathCost(double p);  //set the current path cost of this routing node
  void addto_faninList(routingEdge* e);  //add a routing edge to the fanin list   
  void addto_fanoutList(routingEdge* e);  //add a routing edge to the fanout list   
  void create_faninListItr();  //create a new iterator for faninList
  void remove_faninListItr();  //remove an already existing faninList iterator 
  void create_fanoutListItr();  //create a new iterator for fanoutList
  void remove_fanoutListItr();  //remove an already existing fanoutList iterator 

  double returnCost(int iter,bool pathfinder_pshare);  //get the current cost of this routing node
  char* get_name();  //get the name of this routing node
  int get_type();  //get type of routing node
  double get_R();  //get the resistance of wire routing node
  double get_Cin();  //get the input cap of wire routing node
  double get_Cout();  //get the output cap of wire routing node
  routingNode* get_predecessor();  // get this node's wavefront predecessor

  routingEdge* getnext_faninList();  //get next element from faninList
  routingEdge* getnext_fanoutList();  //get next element from fanoutList
  double get_baseCost();  //get base cost of this routing node
  double get_historyCost(int iter);  //get current history cost of this routing node
  double get_presentSharing(int iter,bool pathfinder_pshare);  //get current present sharing cost of routing node
  int get_occupancy();  //get the current occupancy of this node
  int get_capacity();  //get the capacity of this node
  bool get_isExplored();  //get whether this node has been explored by a wavefront
  double get_pathCost();  //get current path cost of this routing node
  bool foundTerminal(Net* net);  //applicable only to sink routing nodes. if a sink
                                 //terminal of net is mapped to this routing node return true
  void print_info();  //print details about this routing node
  void clear_terminalList();  //clear the list of terminals mapped to this routing node
  void set_counted(bool cn);  //whether this node's sharing has already been included in 
                                //cost calculations 
  bool get_counted();  //true if this node's sharing has been considered in cost calculations
  void print_LRList();  //print the LRList of this routing node...A* debugging

};

class treeNode{
  treeNode* parent;  //parent of this tree node
  list<treeNode*> childList;  //list of children of this tree node
  list<treeNode*>::iterator* child_itr;  //pointer to childList iterator 
  routingNode* rNode;  //the routing node that this tree node corresponds to
  double pathCost;  //cost of the path from the root of the tree to this tree node 
  int magicNum;
 public:
  treeNode(){
    parent=0;
    rNode=0;
    pathCost=0;
    child_itr=0;
    magicNum=31111;
  }
  treeNode(routingNode* r,double p){
    parent=0;
    rNode=r;rNode->set_inrTree(true);  //set rNode's inrTree member to indicate that rNode
                                       //has been included in the route tree of a net
    pathCost=p;
    child_itr=0;
    magicNum=31111;
  }
  ~treeNode(){
    if(magicNum!=31111)
      ERROR("trying to double delete this node...");
    else
      magicNum=0;
  }

  void set_parent(treeNode* t);  //set the parent node of this tree node
  void set_rNode(routingNode* r);  //set the routing node that this tree node represents
  void set_pathCost(double p);  //set the cost of the path from the root to this tree node
  void addto_childList(treeNode* t);  //add a tree node to childList
  void create_childListItr();  //create a new childList iterator
  void remove_childListItr();  //remove childList iterator

  treeNode* get_parent();  //get parent of this tree node
  routingNode* get_rNode();  //get the routing node that this tree node represents
  double get_pathCost();  //get the cost of the path from the root to this tree node
  treeNode* getnext_childList();  //get the next element in childList
  int get_numChildren();  //get the number of children of this tree node

  //get the iterator that points to the node in childList that holds child_node  
  list<treeNode*>::iterator get_childItr(treeNode* child_node);
  //erase the child node pointed to by ch_itr
  void eraseChild(list<treeNode*>::iterator* ch_itr);

};

#endif


