////////////////////////////////////////////////////////////////////
// Net.h: Class that describes a generic Net. 
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

/****************************************************************
Class definition of a generic net. A net has a single source 
terminal and one or more sink terminals.
****************************************************************/    
#ifndef NET_H
#define NET_H
#include "error.h"
#include "Terminal.h"
#include "routingNode.h"

class treeNode;
class Net;
class routingNode;

extern long int sharingCost;
extern long int actual_sharingCost;
extern long int wireCost;
extern long int actual_wireCost;

/*  double cost_to_target(routingNode*,routingNode*); */

class routeTree{
 public:
  treeNode* root;  //the root node of this route tree
  int magicNum;

  routeTree(){
    root=0;magicNum=32111;
  }
  routeTree(treeNode* rt){
    root=rt;magicNum=32111;
  }
  ~routeTree(){
    if(magicNum!=32111)
      ERROR("trying to double delete tree here...");
    else
      magicNum=0;
  }

  void set_root(treeNode* rt);  //set the root of this tree
  treeNode* get_root();  //get the root of this tree
  void print_info();  //print this tree and info relating to it
  void print_info(FILE* fp);

  void preorderTraverse(treeNode* r,int do_what);  //traverse this tree preorder
  void preorderTraverse(treeNode* r,int do_what,FILE* fp);

  void postorderTraverse(treeNode* r,int do_what);  //traverse this tree postorder
  void clear();  //clean up this routing tree
  treeNode* findNode(treeNode* r,routingNode* rnode);  //find rnode in this tree and return                                                          //relevant tree node
  //make this route tree a copy of the tree rooted at tree_root
  void populateWith(treeNode* tree_root);
  void populateWith(treeNode* tree_root,routingNode* sink_rnode);
  treeNode* makeCopy(treeNode* r);

  //get the root of the subtree that sinks at sink_rnode
  treeNode* get_subTree(routingNode* sink_rnode); 

  //increment the occupancies of the routing nodes that constitute this tree
  void incrementOccupancies(int how);

  //find tree node that points to routing node rnode
  treeNode* getNode(routingNode* rnode);

  // Clean the routing nodes in this tree. This is a post-BFS reinit operation.
  void clean_rNodes(treeNode* r);

};

class Terminal;

class Net{
 public:
  char name [MAX_STR_LEN];
  Terminal* source;
  list<Terminal*> sinkList;
  vector<Terminal*>* sinkVector;  //pointer to vector of logic blocks in this netlist
  int numSinks;
  routeTree rTree;
  bool isCounted;
  int magicNum;

  Net(){
    magicNum=17111;
    source=0;
    numSinks=0;
    isCounted=false;
    sinkVector=0;
  }

  ~Net(){
    if(magicNum!=17111)
      ERROR("trying to double delete a node...");
    else
      magicNum=0;
  }

  void set_name(char* n);  //give this net a name
  void set_source(Terminal* s);  //set this net's source terminal
  void addto_sinkList(Terminal* s);  //add a terminal to this net's sink list
  void init_rTreeRoot();  //make the source routing node the root of rTree
  void rip_rTree();  //rip up the existing routing tree, and decrement sharing

  //get a pointer to this net's route tree
  routeTree* get_rTree();

  //rip up the branch of the current route tree that contains sink_rnode
  void rip_rTreeBranch(routingNode* sink_rnode);

  void set_isCounted(bool c);  //whether this net's cost contributions have already
                                 //been considered during an annealing move

  char* get_name();  //get name of this net
  Terminal* get_source();  //get this net's source terminal
  int get_numSinks();  //get num sinks of this net
  treeNode* find_treeNode(routingNode* rnode);  //return tree node that holds the routing
                                                //node rnode
  bool get_isCounted();  //get whether this net's cost contributions have already
                         //been considered during an annealing move
  void print_Info();  //print details about this net
  void print_Info(FILE* fp);

  void print_rTree();  //print the current route tree of this net
  void print_rTree(FILE* fp);

  int get_rNodeCount();  //get the total number of nodes in the current route tree

  //count number of nodes in the current route tree
  void countNodes(treeNode* r,int* count_ptr);
  
  //calculate the sharing and wire cost of this net's current route
  void calcCosts(int* sh_cost,int* wr_cost); 
  
  //recursively calculate wire and sharing cost
  void wsCostCalc(treeNode* r,int* sh_cost,int* wr_cost);

  //reconnect the sub-tree rooted at join_node to this net's current route tree.
  //increment occupancies of the routing nodes that constitute the subtree.
  int reconnect(treeNode* join_node);

  //increment occupancies of the routing nodes that constitute sub-tree rooted at r
  void incrementOccupancies(treeNode* r,int how);

  //Reset the routing nodes in the tree to not counted in sharing calculations
  void reset_counted();

  //initial creation of vector of sinks
  void create_sinkVector();

  //sort vector of sinks in increasing order of separation from source...A*
  void sort_sinkVector();

};

#endif


