////////////////////////////////////////////////////////////////////
// hsraMain.h: Globals and function declarations for hsraMain.y.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

#ifndef BLIF_PARSER_H
#define BLIF_PARSER_H

#include "blifNetlist.h"
#include "hsraArch.h"

/* curr_tok holds the current token type that is being parsed */
int curr_tok=0;

/* pinlist_tok holds the token type for which the pinlist is currently being read */
int pinlist_tok=0;

/* stringList is a list of strings (generally net names, 'open' and digits) */
list<char*> stringList;
list<char*>::iterator stringListItr;

/* declare an blif_netlist */
blifNetlist blif_netlist;

/* declare a generic netlist */
Netlist netlist;

/* declare an HSRA FPGA */
hsraFPGA* FPGA;

/* declare the random seed for this run */
long int randomSeed=0;

/* this is the variable that will hold the current cost of the placement. the cost of a */
/* placement will be defined in terms of sharingCost and wireCost. */ 
double placementCost=0;

/* declare the variable that will hold the total sharing summed across all routing nodes */
/* in the FPGA. this is a cumulative measure of the extent to which the routing nodes are */
/* currently being shared */
long int sharingCost=0;
long int actual_sharingCost=0;

/* declare the variable that will hold the sum of the routing nodes used by the current */
/* route tree of each net in the netlist. this is a measure of the total wirelength of */
/* the current placement. */
long int wireCost=0;
long int actual_wireCost=0;

/* the number of logic blocks + IO pads in netlist */
long int numBlocks=0;

/* the number of nets in netlist */
long int numNets=0;

/* the current temperature of the anneal */
float annealTemp=0.;

/* the size of the move window */
float dLimit=0.;

/* relative importance of sharing factor */
float SHARE_FAC=2.0;

float WIRE_FAC=1.0;

float INNER_NUM=5.0;

float ASTAR_SANITY_SAMPLE = 0.1;

/* direction factor ala VPR */
float ALPHA=1.0;

/* The routing node that is currently being read in from rr_graph.echo */
routingNode* G_currNode=0;

extern vector<vector<double>*>* AstarTable;

bool READ_PLACEMENT=false;

int numClusters = 0;

/* define an entry of the priority queue */
class pqNode{
 public:
  routingNode* rNode;  //the routing node that this priority queue entry represents
  double pathCost;  //the cost of the path from the source to rNode
  routingNode* predNode;  //the fanin routing node that was used to explore rNode
  double heurEstimate;
  int magicNum;

  pqNode(){
    rNode=0;
    pathCost=heurEstimate=HUGE_NUMBER;
    predNode=0;
    magicNum=33111;
  }
  pqNode(routingNode* r,double p,double h,routingNode* pnode){
    if(r==0 || p<0. || h<0.)
      ERROR("error in creating a new PQ node...");
    rNode=r;
    pathCost=p;heurEstimate=h;
    predNode=pnode;
    magicNum=33111;
  }
  ~pqNode(){
    if(magicNum!=33111)
      ERROR("trying to double delete a node...");
    else
      magicNum=0;
  }

  bool operator< (const pqNode& x) const { 
    return (pathCost+heurEstimate) > (x.pathCost+x.heurEstimate);
  }

};

class cachedObj{
 public:
  routingNode* rNode;
  routeTree rTree;
  int clearTree;
  int magicNum;

  cachedObj(){ rNode=0; magicNum=55880; clearTree=0; }
  ~cachedObj(){
    if(magicNum!=55880)
      ERROR("trying to double delete node here...");
    else
      magicNum=0;
  }

};

/***
* FUNCTION DECLARATIONS
***/
void yyerror(char*);
int yylex();
blifNet* find_blifNet(char*);
void print_blifNetlist();
void convert2GenericNetlist(blifNetlist&);
void print_GenericNetlist();


void create_hsraFPGA(int,int,int,float,int);
void initPlacement();
void placeBlock(logicBlock*,hsraBlock*,int);
void printPlacement();
void routeNets(int,bool);
void routeNet(int,Net*,int,bool);
void routeSink(int,Net*,routingNode*,int);
void updatePQ(priority_queue<pqNode>*,treeNode*,routingNode*);
void initPQ(priority_queue<pqNode>&,treeNode*,routingNode*);
treeNode* backTrace(Net*,routingNode*,int,int);
void annealDesign(FILE*);
void generateMove(hsraBlock**,hsraBlock**,int*);
void cacheRoutes(hsraBlock**,hsraBlock**,int*,list<cachedObj*>*,list<cachedObj*>*,list<cachedObj*>*,
                 list<cachedObj*>*);
void undoMove(hsraBlock**,hsraBlock**,int*,list<cachedObj*>*,list<cachedObj*>*,list<cachedObj*>*,
                 list<cachedObj*>*,int);
void makeMove(hsraBlock**,hsraBlock**,int*,int);
void sortSinks(hsraBlock**,hsraBlock**,int*);
void clearCachedRoutes(list<cachedObj*>*,list<cachedObj*>*,list<cachedObj*>*,list<cachedObj*>*,int);
void ripRoutes(logicBlock* lb0,logicBlock* lb1);
void updateHistoryCosts(int);
int calcWireCost();
void printNetsAndRecalcCosts();
void printNetsToFile(FILE*);
void calcStartTemperature();
bool thereIsANeedToAnneal();
void printHSRAPlacement(FILE*);


void processBlifInputs();
void processBlifOutputs();
void processBlifClb();
void processBlifLatch(char*,char*);


extern double cost_to_target(routingNode*,routingNode*);
double flomin(double,double);
bool movable_block(logicBlock*);

vector<char*> *placeVec;
void initExternalPlacement();

void writeOutSourceNode(routingNode*,FILE*);
void writeOutSinkNode(routingNode*,FILE*);
void writeOutInputPinNode(routingNode*,FILE*);
void writeOutOutputPinNode(routingNode*,FILE*);
void writeOutWireNode(routingNode*,FILE*);
void dumpArchInRRGraph(FILE*);

void prepareAstarTable(int,int);
void process_AstarTableEntry(int,char*,double);
void dumpAstarTable();
void dumpClusterAssignments();
void populateAllSinksTables();
void dumpAllSinksTables();
void Dijkstra(priority_queue<pqNode>*,routingNode*,vector<float>*);
void readInClusterAssignments(char*);
void performAstarSanityCheck(FILE*);
void dumpCostToTargetEstimates();
void slogEstimates();
#endif

