////////////////////////////////////////////////////////////////////
// vprArch.h : Classes that describe the VPR architecture.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

/****************************************************************
This file contains class definitions of the VPR FPGA architecture
and the individual clbs that constitute a VPR-style FPGA. Each
vprCLB object holds information about the source/sink/input-pin/
output-pin routing nodes that it provides. Each vprCLB
owns the track segments in the channels that are north and east
of it. Further, the vprCLB also holds a pointer to the logic
block that is currently placed on it.
****************************************************************/    
#ifndef VPR_ARCH_H
#define VPR_ARCH_H
#include "routingEdge.h"
#include "routingNode.h"

extern double flomax(double x,double y);

class vprBaseBlock{
 public:
  int row,col;  //row and col coordinates of this block
  int magicNum;

  vprBaseBlock(int r,int c);
  ~vprBaseBlock();
  int get_row();  //get the row coord of this block
  int get_col();  //get the col coord of this block
  virtual void print_info();  //print details about this block

};


/****************************************************************
This class represents a multi LUT/FF CLB in the VPR architecture.
****************************************************************/
class vprCLB : public vprBaseBlock{
 public:
  routingNode* sourceNode;  //source routing node
  routingNode* sinkNode;  //sink routing node
  list<routingNode*> inputPinList;  //list of input routing nodes
  list<routingNode*> outputPinList;  //list of output routing nodes
  list<routingNode*>::iterator* ipin_itr;  //iterator for inputPinList
  list<routingNode*>::iterator* opin_itr;  //iterator for outputPinList
  routingNode* clock_sinkNode;  //sink routing node for clock routing...NOT USED!!
  logicBlock* currBlock;  //the logic block that is currently placed at this CLB
  int derived_magicNum;

  vprCLB(int r,int c);
  ~vprCLB();

  void set_currBlock(logicBlock* l);  //set the logic block that is placed at this CLB
  logicBlock* get_currBlock();  //get the logic block that is placed at this CLB
  void set_sourceNode(routingNode* r);  //set the source node of this CLB
  void set_sinkNode(routingNode* r);  //set the sink node of this CLB
  routingNode* get_sourceNode();  //get the source node of this CLB
  routingNode* get_sinkNode();  //get the sink node of this CLB
  void addto_inputPinList(routingNode* r);  //add to list of input routing nodes
  void addto_outputPinList(routingNode* r);  //add to list of output routing nodes
  void create_inputPinListItr();  //create an iterator for inputPinList
  void create_outputPinListItr();  //create an iterator for outputPinList
  void remove_inputPinListItr();  //remove the iterator for inputPinList
  void remove_outputPinListItr();  //remove the iterator for outputPinList
  routingNode* getnext_inputPinList();  //get next routing node in inputPinList
  routingNode* getnext_outputPinList();  //get next routing node in outputPinList
  void print_info();  //print details about this CLB

};


/****************************************************************
This class represents an IOB in the VPR architecture. Note that
there are multiple IO elements in a single IOB, and hence there
are multiple source and sink nodes in each IOB.
****************************************************************/
class vprIOB : public vprBaseBlock{
 public:
  vector<routingNode*>* sourceArray;  //ptr to array of source routing nodes
  vector<routingNode*>* sinkArray;  //ptr to array of sink routing nodes
  vector<routingNode*>* inputPinArray;  //ptr to array of input routing nodes
  vector<routingNode*>* outputPinArray;  //ptr to array of output routing nodes
  vector<logicBlock*>* currBlockArray;  //ptr to array of logic blocks placed at this IOB
  int numIOs;  //number of IOs per IOB
  int derived_magicNum;

  vprIOB(int r,int c);
  vprIOB(int r,int c,int n);
  ~vprIOB();
  void addto_sourceArray(routingNode* r,int i);  //add source node to sourceArray at index 'i'
  void addto_sinkArray(routingNode* r,int i);  //add sink node to sinkArray at index 'i'
  void addto_inputPinArray(routingNode* r,int i);  //add input node to inputPinArray at index 'i'
  void addto_outputPinArray(routingNode* r,int i);  //add output node to outputPinArray at index 'i'
  void addto_currBlockArray(logicBlock* l,int i);  //place logic block at 'ith' IO location 
  routingNode* getfrom_sourceArray(int i);  //get node in sourceArray at index 'i'
  routingNode* getfrom_sinkArray(int i);  //get node in sinkArray at index 'i'
  routingNode* getfrom_inputPinArray(int i);  //get node in inputPinArray at index 'i'
  routingNode* getfrom_outputPinArray(int i);  //get node in outputPinArray at index 'i'
  logicBlock* getfrom_currBlockArray(int i);  //get logic block placed at 'ith' IO location
  void print_info();  //print details about this IOB
  int get_numIOs(); //get the number of IOs in this IOB
  void lineup_IOs();

};


class vprFPGA{
 public:
  int numRows,numCols;  //num CLB rows and columns respectively
  int numCLBs,numIOBs;  //num CLBs and IOBs in this architecture
  int numIOs_perIOB;  //number of separate IOs per IOB
  vector<vector<void*>*>* blockArray;  //2-D array of CLBs and IOBs
  list<routingNode*> rnodeList;  //List that holds ALL routing nodes in the architecture.
                                 //This list will be populated from the rr_graph.echo file
                                 //dumped by VPR. The routing edges that connect nodes are
                                 //'owned' by the routing nodes at which the edges originate.
  list<routingNode*>::iterator* rnode_itr;  //iterator for list of routing nodes
  int magicNum;
  int numSinkNodes;

  vprFPGA(int num_rows,int num_cols,int num_ios);
  ~vprFPGA();

  void* get_blockAt(int r,int c);  //get block at location c,r
  int get_numRows();  //get num rows in the fpga...not including IO pads
  int get_numCols();  //get num columns in the fpga...not including IO pads
  int get_numCLBs();  //get num CLBs in this architecture
  int get_numIOBs();  //get num IOBs in this architecture
  void addto_rnodeList(routingNode* r);  //add routing nodes to rnodeList
  routingNode* getnext_rnodeList();  //get next routing node in rnodeList
  void create_rnodeListItr();  //create an iterator for rnodeList
  void remove_rnodeListItr();  //remove iterator for rnodeList
  routingNode* found_routingNode(char* rname);  //look for routing node with rname and return

  //count (and optionally print) the total sharing of routing nodes in this FPGA
  long int countSharing(int opt);

  //count (and optionally print) the total occupancy of congested routing nodes in this FPGA
  long int countCongestedNodeOccupancy(int opt);

  //update the history costs of all routing nodes in this FPGA
  void updateHistoryCosts(int iter,bool during_placement);

  void print_info();  //print arch info about this FPGA

  void lineup_IOBs();  //line up IOs in each IOB's vectors 

};



#endif

